
/***************************************************************************

Copyright (c) 1997 University of Mannheim. All Rights Reserved.

Permission to use, copy, modify, and distribute this software and its 
documentation for educational, research and non-profit purposes, 
without fee, and without a written agreement is hereby granted, 
provided that the above copyright notice, this paragraph and the 
following four paragraphs appear in all copies.  

Permission to incorporate this software into commercial products may 
be obtained from the following address:
University of Mannheim
Lehrstuhl fuer Praktische Informatik IV
L15, 16
68161 Mannheim, Germany

IN NO EVENT SHALL THE UNIVERSITY OF MANNHEIM BE LIABLE TO ANY PARTY 
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, 
INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND 
ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF MANNHEIM HAS BEEN ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF MANNHEIM SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER 
IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF MANNHEIM HAS NO OBLIGATIONS 
TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

The University of Mannheim will appreciate receiving any enhancements and 
modifications of this software made by others, and reserves the right to 
include them in future releases. The University also recommends that this 
software not be redistributed by third parties, but be obtained by all 
directly from the University's server, so that all users of the software 
may receive bug notifications and other relevant announcements.

*/

import java.awt.*;
import java.applet.*;
import java.util.Stack;
import java.lang.*;

/**Hauptklasse: Enthaelt nur ein Applet, das Text ausgibt und 
   das Fenster aufruft, indem die eigentliche Animation ablauft. Zudem
   wird ein ThreadGroup angelegt und weitergegeben
   (um Security-Probleme zu verhindern: suspend,resume).*/
public class MainApplet extends Applet implements Runnable {
    /**Image, wird fuer double buffering Technik verwendet.
    Wird in Scrollbar gezeichnet, in der die Animation ablaeuft.*/
    public Image offscreen;
    /**Dummy-Thread*/
    private Thread thread=null;
    /**Wird an alle  Klassen mituebergeben, um Probleme mit suspend und resume zu verhindern.
    Grund: Klassen in Applets duerfen nur Threads manipulieren, die
    der selben ThreadGroup angehoeren wie die public-Appletklasse.*/
    public ThreadGroup threadGroup=null;
    /**Ruft Fenster auf, in dem Animation ablaueft und uebergibt ThreadGroup*/
    public void init() {
      offscreen=createImage(2600,450);
      MainFrame f = new MainFrame(offscreen,this);
    }
    /**Gibt nur Text auf HTML-Seite aus*/
    public void paint(Graphics g) {
      Font f=new Font("Helvetia",Font.BOLD,30);
      g.setFont(f);
      g.drawString("Interaktive AVL-Baum Animation",20,50);
    }
    public void start() {
      if(thread==null){
        thread=new Thread(this);
        threadGroup=thread.getThreadGroup();
        thread.start();
      }
    }
    public void stop() {
       if(thread!=null && thread.isAlive())  thread.stop();
       thread=null;
    }
    //Dummy-Thread
    public void run() {
       thread.suspend();
    }

}
/**Zeichnet einen Binaerbaum in die Offscreen.*/
class Show extends Object {                
   /**Zeiger Klasse MainFrame*/
   protected MainFrame puCl = null;
   /**Textarea aus Klasse MainFrame*/
   protected TextArea textarea;
   protected Graphics offg;
   /**Offscreen aus Klasse MainFrame*/
   protected Image offscreen;
   /**Scrollbar aus Klasse MainFrame*/
   protected ScrollableCanvas scroll;
   /**verkettete Liste mit Daten des Binaerbaumes*/
   protected VerkListe initTree;
   /**Schrift fuer Schluessel der Knoten*/
   protected Font dataFont = new Font("TimesRoman",Font.BOLD,11);
   /**Schrift fuer die Balancefaktoren der Knoten*/
   protected Font balanceFont = new Font("TimesRoman",Font.BOLD,10);

   /**Einfacher Konstruktor. Wird benutzt bei super()-Aufruf*/
   public Show() {};
   /**Konstruktor: Zeichnet einen Binaerbaum. Anhand der Daten,
   die in der uebergebenen verketteten Liste in stehen, wird Baum aufgebaut*/
   public Show(MainFrame publicClass,VerkListe in) {
      puCl = publicClass;
      textarea = puCl.textarea;
      offscreen = puCl.offscreen;
      Graphics offg = offscreen.getGraphics();
      scroll = puCl.scribble; //ScrollableCanvas
      initTree = in;
      clear(offg);
      printTree(offg,initTree,Color.black,Color.red);
      scroll.sentImage(offscreen);
   }
   /**Konstruktor, der einen Binaerbaum zeichnet. Anhand der Daten,
   die in der uebergebenen verketteten Liste in stehen. Zusaetzlich
   wird der Knoten, mit dem Schluessel data mit der Farbe color
   hervorgehoben*/
   public Show(MainFrame publicClass,VerkListe in, int data,Color color) {
      puCl = publicClass;
      textarea = puCl.textarea;
      offscreen = puCl.offscreen;
      Graphics offg = offscreen.getGraphics();
      scroll = puCl.scribble; //ScrollableCanvas
      initTree = in;
      initTree.search(data);
      printCircle(offg,initTree.Getx1(),initTree.Gety1(),color);
      scroll.sentImage(offscreen);
   }
   /**Binaerbaum wird in offscreen gezeichnet*/
   public void printTree(Graphics g,VerkListe stack,
                         Color newNodeColor,Color newBalColor) {
      if (stack != null) {
         printNodeWithData1(g,stack,newNodeColor);
         printBalance(g,stack,newBalColor);
         printLines(g,stack,newNodeColor);
      }
   }   
   /**Knoten des Baumes mit Schluessel werden gezeichnet.
   Uebergebene verkettete Liste wird durchgegangen und alle Elemente
   der Liste werden gezeichnet als Kreis, versetzter Kreis und Schluessel
   in der Farbe color.*/
   public void printNodeWithData1(Graphics g,VerkListe stack,Color color) {
      stack.GoFirst();
      g.setColor(color);
      g.setFont(dataFont);
      while (! stack.OutOfRange()) {
         int x = stack.Getx1();
         int y = stack.Gety1();
         g.drawOval(x,y,30,30);
         int data = stack.GetData();
         int shift = 8;
         if (data < 10) shift = 13;
            else if ( data < 100) shift = 10;
         g.drawString(""+stack.GetData(),x+shift,y+23);
         g.drawLine(x+4,y+11,x+26,y+11);
         g.drawArc(x+3,y,30,30,90,-180);
         stack.GoNext();
      }
   }
   /**Zeichnet einen Kreis und versetzten Kreis.
   In Farbe color und Position x,y. Dient dazu einen Knoten hervorzuheben.
   Wird benutzt, wenn Mauszeiger ueber Knoten sich befindet.*/
   public void printCircle(Graphics g,int x,int y,Color color) {
      g.setColor(color);
      g.drawOval(x,y,30,30);
      g.drawLine(x+4,y+11,x+26,y+11);
      g.drawArc(x+3,y,30,30,90,-180);
   }
   /**Zeichnet in den Konten(dargestellt als Kreis) den Balancefaktor
   in das obere Drittel in der Farbe color. Dies geschieht mit allen
   Elementen der verketteten Liste.*/
   public void printBalance(Graphics g,VerkListe stack, Color color) {
      stack.GoFirst();
      g.setColor(color);
      g.setFont(balanceFont);     
      while (! stack.OutOfRange()) {
         int x = stack.Getx1();
         int y = stack.Gety1();
         int bal = stack.GetBalance();
         String string ="";
         switch(bal) {
            case 0: string = " 0";break;
            case 1: string = "+1";break;
            case 2: string = "+2";break;
            case -1: string = "-1";break;
            case -2: string = "-2";break;
         }
         g.drawString(string,x+11,y+10);
         stack.GoNext();
      }
   }
   /**Zeichnet Linien, die die Knoten des Binaerbaumes verbinden.
   Anhand der verketten Liste wird der Abstand zwischen den Knoten
   ermittelt und dann in der Farbe color gezeichnet.*/
   public void printLines(Graphics g,VerkListe stack,Color color) {
      g.setColor(color);
      int treeHeight = puCl.Tree.CurrentHeight;
      stack.GoFirst();
      while (!stack.OutOfRange()) {
         int x = stack.Getx1();
         int y = stack.Gety1();
         int restTreeHeight = treeHeight - puCl.Tree.getHeight(y);
         restTreeHeight = puCl.Tree.xHoch2(restTreeHeight);
         int left = stack.GetLeft();
         int right= stack.GetRight();
         if (left>0) g.drawLine(x+4,y+25,x-(restTreeHeight*10)+15,y+60);
         if (right>0) g.drawLine(x+26,y+25,x+(restTreeHeight*10)+15,y+60);
         stack.GoNext();
      }
   }
   /**Loescht die offscreen*/
   public void clear(Graphics g) {
      g.clearRect(0,0,offscreen.getWidth(null),offscreen.getHeight(null));
   }
}

/**Wird aufgerufen, wenn ein neues Element in den AVL-Baum eingefuegt wird.
Alle Bewegungen der Animation die durch das Einfuegen entstehen
(blaue Linie nachziehen beim Einfuegen,u.U. Rotation(en) mit blauem Pfeil,
blaue Linie wieder neutraliseren, Textausgabe in Textarea) werden hier abgehandelt.
Nach jeder Bewegung wird gesamter Baum wieder neu aufgebaut.*/
class Move extends Show implements Runnable {
   protected Thread thread;
   /**AVL-Baum-Daten vor Einfuegen sind hier abgelegt*/
   protected VerkListe oldTree;
   /**AVL-Baum-Daten nach Einfuegen des neuen Elements sind hier abgelegt (aber noch vor Rotation)*/
   protected VerkListe initTree;
   /**AVL-Baum-Daten nach erster Rotation sind hier abgelegt, u.U. auch Leer*/
   protected VerkListe rot1Tree;
   /**AVL-Baum-Daten nach zweiter Rotation sind hier abgelegt u.U. auch Leer*/
   protected VerkListe rot2Tree;
   /**Einfuegepfad des neue Elements wurde beim Einfuegen festgehalten*/
   protected VerkListe einfPfad;
   /**Bewegungsdaten der Knoten waehrend der 1. Rotation werden hier abgelegt*/
   protected VerkListe moveNode1 = null;
   /**Bewegungsdaten der Knoten waehrend der 2. Rotation werden hier abgelegt*/
   protected VerkListe moveNode2 = null;
   /**Bewegungsdaten des zu einfuegenden Knotens werden hier abgelegt*/
   protected VerkListe moveInsert = null;
   /**Zeigt an, ob eine Bewegung aktiv ist*/
   protected boolean activ;
   /**Zeigt an, ob eine synchrone oder sequentielle Abarbeitung notwendig ist.
   Synchron: Alle Element einer verk. Liste werden gleichzeitig bewegt,
   z.B. Rotation.
   Sequentiell: Zuerst wird 1. Element der verk. Liste zu seinem Endpunkt
   bewegt danach 2. usw, notwendig bei der Einfuegebewegung des neuen Knoten.*/
   protected boolean synchron;
   /**Zeigt an, ob Rotation in Uhrzeigerrichtung.*/
   protected boolean clockwise;
   /**Bewegungsanfang: Bei welchem Element in verk. Liste beginnt Bewegung*/
   protected int begin;
   /**Bewegungsende: Bei welchem Element in verk. Liste endet Bewegung*/
   protected int end;
   /**SleepTime des Threads zwischen den einzelnen Bewegungsvorgaengen*/
   protected int sleepTime;
   /**x-Koordinate des blauen Pfeils bei Rotation
   (=x-Koordinate des Knotens, um den Rotation erfolgt*/
   protected int rotX;
   /**x-Koordinate des blauen Pfeils bei Rotation
   (=x-Koordinate des Knotens, um den Rotation erfolgt*/
   protected int rotY;
   /**Farbe des Knotens*/
   protected Color node;
   /**Farbe des Balancefaktors im Knoten*/
   protected Color nodebalance;
   protected Thread threadBackground;
   /**Einfacher Konstruktor. Bei Aufruf als Superklasse.*/
   public Move() {};
   /**Konstruktor. In verk. Liste 1 bis 5 sind AVL-Baum-Daten abgelegt:
   1.Baum vor Einfuegen 2.Baum nach Einfuegen(vor Rotation) 3.Baum nach
   erster Rotation, u.U. leer 4.Baum nach zweiter Rotation u.U. leer
   5.Einfuegepfad des neuen Knoten.*/
   public Move(MainFrame newPublicClass,VerkListe stack1,VerkListe stack2,
               VerkListe stack3,VerkListe stack4,VerkListe stack5,Thread t) {
      puCl = newPublicClass;
      threadBackground = t;
      scroll = puCl.scribble; //ScrollableCanvas
      offscreen = puCl.offscreen; //double-buffering screen
      offg = offscreen.getGraphics();
      textarea = puCl.textarea;   //Textarea for output
      sleepTime = puCl.sleepTime;
      oldTree = stack1;
      initTree = stack2;
      rot1Tree = stack3;
      rot2Tree = stack4;
      einfPfad = stack5;
      adoptInfos(oldTree,initTree); //new Coordinates
      moveInsert = computeMoveInsert(einfPfad);  //Einfgebewegung berechenen
      moveNode1 = computeMoveNode(initTree,rot1Tree); //Rotationsbewegung
      moveNode2 = computeMoveNode(rot1Tree,rot2Tree);
   }

   public void start() {
      thread = new Thread(this);
      thread.start();
   }

   public void stop() {
      thread.stop();
   }
   /**Kontrolliert Einfuegevorgang. Zuerst wird neuer Knoten an die richtige
   Stelle im Baum bewegt. Zugleich wird Einfuegepfad gekennzeichnet. Danach
   wird der Einfuegepfad wieder zurueckgegangen. Bei jedem Knoten kontrolle,
   ob eine Rotation notwendig ist. Gegebenfalls wird notwendige Rotation
   durchgefuehrt. Permanent wird im Textfenster beschrieben, was im Animationsfenster
   passiert.*/
   public void run() {
      puCl.nextStep.enable();
      begin = 1;end=1;
      synchron = false;
      //Bewegung Knoten einfuegen
      while (! moveReady(moveInsert)) {
         try {
            printNodeWithData2(moveInsert,begin,end,oldTree,Color.black,Color.red);
            nextCoordinate(moveInsert,begin,end);
            thread.sleep(sleepTime);
            stopOrRun();
         }
         catch (InterruptedException e) {}
      }
      //Einfuegepfad wieder zurueck
      moveInsertReverse();
      if (moveInsert != null) {
         begin =1;end = 1;
         int oldBegin = 0;
         activ = true;
         while( !moveReady(moveInsert) && activ) {
            try {
               printLinesBack(moveInsert,initTree,begin,end);
               nextCoordinate(moveInsert,begin,end);
               if (begin > oldBegin) {
                  oldBegin = begin;
                  correctBalance();
               }
               thread.sleep(sleepTime);
               stopOrRun();
            }
            catch(InterruptedException e) {}
         }
         if (activ) {
            begin++;end++;correctBalance();
            clear (offg);
            printTree(offg,initTree,Color.black,Color.red);
            scroll.sentImage(offscreen);
         }
      }
      else {
         clear (offg);
         printTree(offg,initTree,Color.black,Color.red);
         scroll.sentImage(offscreen);
      }
      if (moveNode1 != null) {
         begin = 1;end=moveNode1.Length();
         synchron = true;
         findRotCoord(moveNode1);
         while (! moveReady(moveNode1)) {
            try {
               printNodeWithData2(moveNode1,begin,end,initTree,Color.gray,Color.gray);
               nextCoordinate(moveNode1,begin,end);
               thread.sleep(sleepTime);
               stopOrRun();
            }
            catch (InterruptedException e) {}
         }
         clear (offg);
         printTree(offg,rot1Tree,Color.black,Color.red);
         scroll.sentImage(offscreen);
      }
      if (moveNode2 != null) {
         begin = 1;end=moveNode2.Length();
         synchron = true;
         findRotCoord(moveNode2);
         while (! moveReady(moveNode2)) {
            try {
               printNodeWithData2(moveNode2,begin,end,rot1Tree,Color.gray,Color.gray);
               nextCoordinate(moveNode2,begin,end);
               thread.sleep(sleepTime);
               stopOrRun();
            }
            catch (InterruptedException e) {}
         }
         clear (offg);
         printTree(offg,rot2Tree,Color.black,Color.red);
         scroll.sentImage(offscreen);
      }
      offg.dispose();
      puCl.einfuegenOk.enable();
      puCl.loeschenOk.enable();
      if (threadBackground != null) threadBackground.resume();
   }
   /**Zeichnet AVL-Baum. Entweder mit markierten Einfuegepfad oder mit blauem
   Pfeil(zeigt Rotationsrichtung an), falls gerade Rotation stattfindet.
   Bewegungsdaten in stack, Beginn bei Element b in stack, analog Bewegungsende e,
   AVL-Baumdaten tree(im Hintergrund bei Rotation),Knotenfarbe newNode,
   Balancefaktorfarbe newBal.*/
   public void printNodeWithData2(VerkListe stack,int b,int e,
                                  VerkListe tree,Color newNode,Color newBal){
      if (stack != null) {
         clear(offg);
         printTree(offg,tree,newNode,newBal);
         offg.setFont(dataFont);
         offg.setColor(Color.blue);
         for(int i = b;i<=e; i++) {
            stack.Go(i);
            int x = stack.Getx1();
            int y = new Float(stack.Gety2()).intValue();
            int data = stack.GetData();
            offg.drawOval(x,y,30,30);
            int shift = 8;
            if (data < 10) shift = 13;
               else if ( data < 100) shift = 10;
            offg.drawString(""+data,x+shift,y+23);
            offg.drawLine(x+4,y+11,x+26,y+11);
            offg.drawArc(x+3,y,30,30,90,-180);            
            if(b==e) scroll.setScrollbarValues(x,y);
         }
         if (!synchron) {  //print thick lines if not synchron
            for(int k = 1;k<=e; k++) {
               stack.Go(k);
               printThickLines(stack,true);
            }
         }
         if (synchron) drawArrow();
         scroll.sentImage(offscreen);
      }
   }
   /**Berechnet Koordinaten der naechsten Bewegung. Die Bewegungsdaten stehen
   in stack. Im stack stehen die Anfangs- und Endkoordinaten jedes Knoten der
   bewegt werden soll. Die Bewegungsdaten koennen entweder synchron oder
   sequentiell bearbeitet werden. Bei synchroner Bewegung werden alle Elemente
   gleichzeitig von ihrem Anfangspunkt bis zu ihrem Endpunkt bewegt, d.h. bei
   allen Elementen von stack wird naechster Punkt berechnet. Bei sequentieller
   Abarbeitung wird zuerst 1. Element zu ihrem Endpunkt bewegt
   danach 2.,3. usw., d.h. immer nur bei aktuellem Elememt wird naechste
   Koordinate berechnet.*/  
   public void nextCoordinate(VerkListe stack,int b,int e) {
      for(int i = b; i<=e; i++) {
         stack.Go(i);
         int steps = stack.GetBalance(); //zu verbleibende x-Koordin.
         int direction = stack.GetRight(); //x-Richtung
         int stepsMinus = Math.abs(direction);
         float y;
         float m = stack.GetM();
         if (stepsMinus == 0) stepsMinus = Math.abs(new Float(m).intValue());
         if (steps>0) {
            if (steps-stepsMinus<0) {
               if (direction < 0) direction = -steps;
                  else direction = steps;
               //steps = 0;
            }
            steps = steps - stepsMinus;
            if (direction != 0) {
               int x = stack.Getx1() + direction;
               stack.Setx1(x);
               y = stack.Gety2() + Math.abs(direction)*m;
            }
            else {
               if (steps-stepsMinus<0) {
                  if (m < 0) m  = -steps;
                     else m = steps;
               //steps = 0;
               }
               y = stack.Gety2() + m;
            }
            stack.SetBalance(steps);
            stack.Sety2(y);
         }
         else {if (!synchron) {
                  stack.SetBalance(-1);
                  if (begin < stack.Length()) {
                    begin++;end++;
                    nextCoordinate(stack,begin,end);
                  }
               }  else {stack.SetBalance(-1);}
         }

      }      
   }

   /**Ueberprueft, ob alle Element in Bewegungsliste stack ihren Endpunkt erreicht haben.*/
   public boolean moveReady (VerkListe stack) {
      if (stack != null) {
         boolean ready = true;
         stack.GoFirst();
         while ( ! stack.OutOfRange() &&  ready) {
            if (stack.GetBalance() >= 0 ) {ready = false;}
             else {stack.GoNext();}
         }
         return ready;
      }
      else {return true;}
   }
   /**Zeichnet von von aktuellem Ort bis zur Wurzel im Baum ein dicke blaue Linie.*/
   public void printLinesBack(VerkListe stack,VerkListe baum,int b,int e ) {
      clear(offg);
      printTree(offg,baum,Color.black,Color.red);
      stack.GoFirst();
      while ( ! stack.OutOfRange()) {
         printThickLines(stack,false);
         stack.GoNext();
         //System.out.println("begin end" + begin);
      }
      stack.Go(begin);
      int y = new Float(stack.Gety2()).intValue();
      scroll.setScrollbarValues(stack.Getx1(),y);
      //System.out.println("----------------------");
      scroll.sentImage(offscreen);
   }
   /**Die Einfuegebewegung von Wurzel bis zum Endpunkt des neuen Knoten wird umgekehrt.
   Beim Einfuegen wird blaue Linie von Wurzel bis Endpunkt des neuen Knoten
   nachgezogen. Nun wird diese Bewegung umgekehrt und es kann eine Bewegung
   von Endpunkt des neu eingefuegten Knoten bis Wurzel simuliert werden.*/
   public void moveInsertReverse() {
      boolean found = false;
      VerkListe tmp = new VerkListe();
      int laenge = moveInsert.Length();
      for(int i = laenge;i>=1;i--) {
         moveInsert.Go(i);
         if (moveInsert.GetLeft()!=0) {
            found = true;
            //int yint = stackmove.Gety1();
            //float yfloat = (new Integer(yint).floatValue());
            //int yint =(new Float(stackmove.Gety2()).intValue());
            int x1 = moveInsert.GetLeft();
            int x2 = moveInsert.Getx1();
            int steps = Math.abs(x1-x2);
            tmp.PushToEnd(moveInsert.GetData(),steps,moveInsert.Getx1(),
            moveInsert.Gety2(),0-moveInsert.GetM(),0-moveInsert.GetRight());
            tmp.Sety1(moveInsert.Gety1());
            tmp.SetLeft(moveInsert.GetLeft());
         }
      }
      if (found) moveInsert=tmp;
         else moveInsert = null;
   }
   /**Zeichnet eine dicke blaue Linie von Wurzel bis zum Endpunkt des neu eingefuegten Knotens.*/
   public void printThickLines(VerkListe stack,boolean direction) {
      offg.setColor(Color.blue);
      if (stack.GetLeft() != 0) {
         int aa = 0;
         int bb = 0;
         if (stack.GetRight() > 0) {aa=-1;bb=1;}
            else {aa = 1;bb=-1;
         }
         if (direction) {aa = aa * (-1); bb = bb * (-1);}
         int x1 = stack.GetLeft() + 15;
         int y1 = stack.Gety1();
         int x2 = stack.Getx1() + 15;
         int y2 = new Float(stack.Gety2()).intValue();
         int intx[] = new int [4];
         int inty[] = new int [4];
         intx[0] = x1-1;
         inty[0] = y1+aa;
         intx[1] = x1+1;
         inty[1] = y1+bb;
         intx[2] = x2+1;
         inty[2] = y2+bb;
         intx[3] = x2-1;
         inty[3] = y2+aa;
         offg.fillPolygon(intx,inty,4);
         //if(!synchron) scroll.setScrollbarValues(x,y);
         //if(!direction) scroll.setScrollbarValues(x2,y2);
      }
   }
   /**Uebeprueft, ob eine Rotation notwendig ist. Geschieht immer dann, wenn
   beim Zurueckgehen des Einfuegepfads beim Zeichnen ein Knoten erreicht
   wurde. Beim Erreichen eines Knoten wird immer Ergebnis der Untersuchung
   im Textfenster ausgegeben.*/
   public void correctBalance() {
      int i = 0;
      int data = 0;
      int laenge = 0;
      laenge = einfPfad.Length();
      i = laenge - begin + 1;
      einfPfad.Go(i);
      data = einfPfad.GetData();
      if (i==laenge) {
         printLinesBack(moveInsert,initTree,begin,end);
         //text.setText("");
         textarea.appendText("Baum wieder aufwaerts entlang des Einfuegepfads ausbalancieren"+"\n");
         textarea.appendText("Gegebenfalls rotieren (falls nicht |Balancefaktor|<= 1) und Vorgang beenden"+"\n");
         textarea.appendText("oder bis zur Wurzel Vorgang fortsetzen."+"\n");
         textarea.appendText(begin+".  ");
         textarea.appendText("Neuer Knoten "+data+" ist ein Blatt des Baumes."+"\n");
         textarea.appendText("Knoten bekommt Balancefactor 0."+"\n");
      }
      else {
         int hRechts = einfPfad.GetRight();
         int hLinks = einfPfad.GetLeft();
         int balance = hRechts - hLinks;
         initTree.search(data); 
         initTree.SetBalance(balance);
         textarea.appendText(begin+".  ");
         textarea.appendText("Untersuche Knoten "+data+"\n");
         textarea.appendText("Hoehe rechter Teilbaum = "+hRechts+"\n");
         textarea.appendText("Hoehe linker Teilbaum  = "+hLinks+"\n");
         textarea.appendText("Neuer Balancefaktor = "+balance);
         if (i > 1) { // ?pal(p)=0 && pal(Sohn(p))=0  dann fertig
            einfPfad.Go(i-1);
            if(einfPfad.GetLeft()<0 && einfPfad.GetRight()<0) {
               //einfPfad.Go(i+1);
               //int nextData = einfPfad.GetData();
               //int nextBalance = einfPfad.GetRight()-einfPfad.GetLeft();
               if (Math.abs(balance)!=2) {
                  //System.out.println("fertig");
                  textarea.appendText("\n");
                  textarea.appendText("Teilbaum-Hoehe des Knoten "+data+"\n");
                  textarea.appendText("hat sich nicht veraendert => fertig"+"\n");
                  clear (offg);
                  printTree(offg,initTree,Color.black,Color.red);
                  scroll.sentImage(offscreen);
                  offg.dispose();
                  puCl.einfuegenOk.enable();
                  puCl.loeschenOk.enable();
                  if(threadBackground != null) threadBackground.resume();
                  thread.stop();
               }
            }
            einfPfad.Go(i);
         }
         if (Math.abs(balance)<=1) {
            textarea.appendText("  => keine Rotation notwendig"+"\n");
            if (i==1) textarea.appendText("\n"+"Wurzel erreicht => fertig\n\n");
         }
         else {
            textarea.appendText(" => Rotation notwendig"+"\n");
            einfPfad.Go(i+1);
            int balancePrev = einfPfad.GetRight()-einfPfad.GetLeft();
            int dataPrev = einfPfad.GetData();
            String rotation = "";
            String reason = "Balance("+data+")="+balance+" und ";
            if (balance==-2) {
               reason = reason + "Balance linker Sohn("+dataPrev+")="+balancePrev;
               if (balancePrev==1)  rotation = "Links/Rechts";           
                  else rotation = "Rechts";
            } else {
               reason = reason + "Balance rechter Sohn("+dataPrev+")="+balancePrev;
               if (balancePrev==1) rotation = "Links";
                  else rotation = "Rechts/Links";
            }
            rotation = "  => "+rotation + "-Rotation";
            textarea.appendText(reason+"\n");
            textarea.appendText(rotation+"\n");
            activ = false;
         }
      }
   }

   /**Aktualisiert AVL-Baum-Daten. Beim Einfuegen kann es vorkommen, dass
   ein Baum sich verbreitert. Dadurch stimmen die Koordinaten des aktuellen
   Baum oldPos nicht mehr.*/
   public void adoptInfos(VerkListe shortStack,VerkListe longStack) {
      if (shortStack != null && longStack != null) {
         shortStack.GoFirst();
         longStack.GoFirst();
         while( !shortStack.OutOfRange() && !longStack.OutOfRange()) {
            if(shortStack.GetData() == longStack.GetData()) {
               shortStack.Setx1(longStack.Getx1());
               shortStack.Sety1(longStack.Gety1());
               shortStack.SetBalance(longStack.GetBalance());
               shortStack.GoNext();
               longStack.GoNext();
            }
            else { longStack.GoNext();}
         } 
      }
   }
   /**Die Einfuegebewegung wird berechnet. Zuerst wird eine Bewegung 
   von oben senkrecht zur Wurzel eingefuegt. Dann werden die Bewegungen von Knoten
   zu Knoten des Baumes entlang des Einfuegepfads bis zum Endpunkt ermittelt.*/
   public VerkListe computeMoveInsert(VerkListe stack) {
      VerkListe tmp = new VerkListe();
      if (stack != null) {
         stack.GoLast();
         int data = stack.GetData();
         float yfloat = -30; float mfloat =5;
         tmp.PushToEnd(data,80,1300,yfloat,mfloat,0);
         int laenge = stack.Length();
         for (int i=1;i<laenge;i++) {
            stack.Go(i);
            Integer y1Int = new Integer(stack.Gety1());
            float y1float = y1Int.floatValue() + 25;
            int startx = 0;
            int x1int = stack.Getx1();
            Integer x1Int = new Integer(x1int);
            float x1float = x1Int.floatValue();
            stack.Go(i+1);
            Integer y2Int = new Integer(stack.Gety1());
            float y2float = y2Int.floatValue();
            int x2int = stack.Getx1();
            Integer x2Int = new Integer(x2int);
            float x2float = x2Int.floatValue();
            int steps = 0;
            int steplength = 0;
            int direction = 0;
            float m = 0;
            if (x1int>x2int) {
               x1float = x1float + 4 -15;
               m = (y2float-y1float) / (x2float-x1float);
               m=0-m;
               tmp.PushToEnd(data,5,x1int,y1float,0,-5);
               startx = x1int;
               x1int = x1int + 4 -15;
               steps = Math.abs(x1int-x2int);
               steplength = steps / 5;
               direction = 0-steplength;
            }
            else {
               x1float = x1float + 26 -15;
               m = (y2float-y1float) / (x2float-x1float);
               tmp.PushToEnd(data,5,x1int,y1float,0,5);
               startx = x1int;
               x1int = x1int + 26 - 15;
               steps = Math.abs(x1int-x2int);
               steplength = steps / 5;
               direction = steplength;
            }
            tmp.PushToEnd(data,steps,x1int,y1float,m,direction);
            tmp.SetLeft(x1int);
            tmp.Sety1(new Float(y1float).intValue());
            if (laenge>(i+1)) {
               m=5;
               tmp.PushToEnd(data,25,x2int,y2float,m,0);
            }
         }
      }
      else {
         initTree.GoFirst();
         int data = initTree.GetData();
         float yfloat = -30; float mfloat =5;
         tmp.PushToEnd(data,60,1300,yfloat,mfloat,0);
      }
      return tmp;
   }
   /**Rotationsbewegung wird berechnet. In stack1 stehen alle Koordinaten
   der Knoten vor Rotation in stack2 die danach. Wenn sich die Koordinate
   eines Knoten sich veraendert hat wird die Bewegung von Koordinate aus
   stack1 (Anfangspunkt) nach Koordinate aus stack2 (Endpunkt) ermittelt
   und in der Bewegungsliste abgelegt, die zurueckgegeben wird.*/
   public VerkListe computeMoveNode(VerkListe stack1,VerkListe stack2) {
      VerkListe tmp = null;
      if (stack1 != null && stack2 != null) {
         tmp = new VerkListe();
         stack1.GoFirst();
         stack2.GoFirst();
         while( !stack1.OutOfRange()){
            //System.out.println("d"+stack1.GetData()+"d"+stack2.GetData());
            if (stack1.Getx1()!=stack2.Getx1() || stack1.Gety1()!=stack2.Gety1()) {
               Integer y1Int = new Integer(stack1.Gety1());
               float y1 = y1Int.floatValue();
               Integer x1Int = new Integer(stack1.Getx1());
               float x1 = x1Int.floatValue();
               Integer y2Int = new Integer(stack2.Gety1());
               float y2 = y2Int.floatValue();
               Integer x2Int = new Integer(stack2.Getx1());
               float x2 = x2Int.floatValue();
               float m = (y2-y1) / (x2-x1);
               int steps = Math.abs(stack1.Getx1()-stack2.Getx1());
               int steplength = steps / 10;
               int direction=0;
               if (stack1.Getx1()>stack2.Getx1()) {direction = 0-steplength;m=0-m;}
               else {direction = steplength;}
               tmp.PushToEnd(stack1.GetData(),steps,stack1.Getx1(),y1,m,direction);
               //System.out.println("x1"+x1+"y1"+y1+"x2"+x2+"y2"+y2);
            }         
            stack1.GoNext();
            stack2.GoNext();
         }
      }
      return tmp;
   }
   /**Zeichnet einen bleien Pfeil. Klassenvariablen rotX,rotY geben Ort an
   clockwise zeigt richtung an. Soll bei einer Rotation zus. Richtung
   der Rotation verdeutlichen.*/
   public void drawArrow() {
      int x = rotX - 38;
      int y = rotY - 20;
      int w = 90;
      int h = 60;
      offg.setColor(Color.blue);
      for(int i=0;i<=11;i++) {offg.drawArc(x+i,y,w,h,0,180);}
   
      int intx[] = new int [3];
      int inty[] = new int [3];
      if (clockwise)  x += w;
      y += 30;
      intx[0] = x -5;
      inty[0] = y;
      intx[1] = x+11+5;
      inty[1] = y;
      intx[2] = x+6;
      inty[2] = y+10;
      offg.fillPolygon(intx,inty,3);
   }
   /**Ermittelt Element in stack, um das die Rotation stattfindet. Wird
   von drawArrow() benoetigt, um Koordinaten des blauen Pfeils herauszufinden.*/
   public void findRotCoord(VerkListe stack) {
      if (stack != null) {
         rotY = 800;
         stack.GoFirst();
         while ( !stack.OutOfRange() ) {
            int y = new Float(stack.Gety2()).intValue();
            if (y<rotY) {
               rotY = y;
               rotX = stack.Getx1();
               if (stack.GetRight() > 0) clockwise = true;
                  else clockwise = false;
            }
            stack.GoNext();
         }
      }      
   }
   /**Kontrolliert den Animationablauf.
   Falls Button Stop aktiviert wurde ruht die Animation, bei Button Next
   geht Animation zum naechsten Bewegungsschritt ueber und haelt wieder an,
   bei Run lauft Animation weiter.*/
   public void stopOrRun() {
     if(puCl.nextStep.isEnabled()) {
       while ( puCl.run.isEnabled()) {
          try {
             thread.sleep(5);
          }
          catch(InterruptedException e) {}
       }
     }
     else {
       //puCl.run.enable();
       puCl.nextStep.enable();
       //puCl.stop.disable();
       //puCl.stopRunCardLayout.show(stopRunPanel,"run");
       while (puCl.run.isEnabled() && puCl.nextStep.isEnabled()) {
          try {
             thread.sleep(5);
          }
          catch(InterruptedException e) {}
       }
     }
   }    

}
/**Wird aufgerufen, wenn ein Element aus dem AVL-Baum geloescht wird.
Alle Bewegungen der Animation die durch das Loeschen entstehen
(blaue Linie nachziehen beim Suche des Elements ,u.U. Rotation(en) mit blauem Pfeil,
blaue Linie wieder neutraliseren, Textausgabe in Textarea) werden hier abgehandelt.
Nach jeder Bewegung wird gesamter Baum wieder neu aufgebaut.*/
class MoveDelete extends Move implements Runnable {
   /**Enthaelt mehrere verschiedene verk. Listen, in den AVL-Baum-Daten abgelegt sind.
   Die verk. Listen sind durch ein Erkennungszeichen getrennt, um diese spaeter wieder
   trennen zu koennen. Dieses Vorgehensweise ist notwendig, weil ,im
   Gegensatz zum Einfuegen, u.U. mehrer Einfach- oder Doppelrotationen notwendig
   sein koennen, um Baum auszubalancieren nach Loeschen. Da nach jeder Rotation eine
   neue verk. Liste erstellt werden muss, ist Anzahl verk. Listen ist bekannt.*/
   private VerkListe allMoves;
   boolean forward = true;
   /**Zeigt an, ob ein innerer Knoten im AVL-Baum geloescht wurde*/
   boolean innererKnoten = false;  //wurde ein innerer Knoten geloescht ?
   /**Zu loeschender Schluessel*/
   int del;
   /**Zeigt an, welche Rotationsaktion durchegefuehrt wurde.
   Steht am Kopf jeder verk. Liste.0=keine/1=L/2=R/3=LR/4=RL - Rotation */
   int action;
   /**Konstruktor. In stack1 stehen alle verk. Liste die bei Rotationen
   entstanden sind. Der stack2 enthaelt die Knoten des Suchpfads
   von Wurzel bis zum geloeschten Schluessel.*/
   public MoveDelete(MainFrame newPublicClass,VerkListe stack1,VerkListe stack2,
               VerkListe stack3,int newDel,Thread t) {
      super();
      puCl = newPublicClass;
      threadBackground = t;
      scroll = puCl.scribble; //ScrollableCanvas
      offscreen = puCl.offscreen; //double-buffering screen 
      offg = offscreen.getGraphics();
      textarea = puCl.textarea;   //Textarea for output
      sleepTime = puCl.sleepTime;
      oldTree = stack1;
      allMoves = stack2;
      einfPfad = stack3;
      del = newDel;
      if (allMoves!=null) {
         initTree = new VerkListe();
         allMoves.GoFirst();
         action = allMoves.GetBalance();
         allMoves.GoFirst();
         initTree.Split(allMoves);
         puCl.posOld = initTree.Clone();
      }
      //adoptInfos(oldTree,initTree); //new coordinates form old tree
      moveInsert = computeMoveDelete(einfPfad);  //SuchPfad berechenen
      //moveNode1 = computeMoveNode(initTree,rot1Tree);
      //moveNode2 = computeMoveNode(rot1Tree,rot2Tree);
   }

   public void start() {
      thread = new Thread(this);
      thread.start();
   }

   public void stop() {
      thread.stop();
   }
   /**Kontrolliert Loeschvorgang. Zuerst wird zu loeschender Knoten gesucht.
   Zugleich wird Suchpfad gekennzeichnet. Danach
   wird der Suchpfad wieder zurueckgegangen. Bei jedem Knoten kontrolle,
   ob eine Rotation notwendig ist. Gegebenfalls wird notwendige Rotation
   durchgefuehrt. Permanent wird im Textfenster beschrieben, was im Animationsfenster
   passiert.*/
   public void run() {
      puCl.nextStep.enable();
      if (moveInsert != null) {
         begin = 1;end=1;
         int oldBegin =1;
         synchron = false;
         while (! moveReady(moveInsert)) {
            try {
               if(begin==oldBegin) {
                  oldBegin++;
                  suchVorgang();
               }
               printNodeWithData2(moveInsert,null,begin,end,oldTree,Color.black,Color.red);
               nextCoordinate(moveInsert,begin,end);
               thread.sleep(sleepTime);
               stopOrRun();
            }
            catch (InterruptedException e) {}
         }
         begin++;end++;
         suchVorgang();
         textarea.appendText("\n");
         textarea.appendText("Baum wieder aufwaerts entlang des Suchpfads ausbalancieren"+"\n");
         textarea.appendText("Gegebenfalls rotieren (falls nicht |Balancefaktor|<= 1)."+"\n");
         textarea.appendText("Solange bis: bal(x)= 0 wird zu +/-1   oder  Wurzel erreicht."+"\n"+"\n");
         moveInsertReverse();
         if (moveInsert != null) {
            begin =1;end = 1;
            if(innererKnoten) oldBegin = 1;
               else oldBegin = 2;
            activ = true;            
            while( !moveReady(moveInsert))  {
               try {
                  if (begin == oldBegin) {
                     oldBegin++;
                     correctBalance();
                     //einfPfad.Go(begin);
                     //System.out.println(begin+". data"+einfPfad.GetData());
                  }
                  printLinesBack(moveInsert,oldTree,begin,end);
                  nextCoordinate(moveInsert,begin,end);
                  thread.sleep(sleepTime);
                  stopOrRun();
               }
               catch(InterruptedException e) {}
            }
            if (einfPfad.Length()>0) {
               correctBalance();
               clear (offg);
               printTree(offg,oldTree,Color.black,Color.red);
               scroll.sentImage(offscreen);
            }
            /*if (activ) {begin++;end++;correctBalance();}
            clear (offg);
            printTree(offg,initTree,Color.black,Color.red);
            scroll.sentImage(offscreen); */
         }
         else {
            clear (offg);
            printTree(offg,oldTree,Color.black,Color.red);
            scroll.sentImage(offscreen);
         }
      }
      offg.dispose();
      puCl.einfuegenOk.enable();
      puCl.loeschenOk.enable();
      if (threadBackground !=null) threadBackground.resume();
   }
   /**Untersucht welche Rotation durchzufuehren ist. Veranlasst und zeigt an,
   welche Schritte notwendig sind.*/
   public void rotation(int action) {
      if (allMoves.Length() > 0) {
      allMoves.GoFirst();
         //action = allMoves.GetBalance();//Gehe SuchPfad bis Wurzel zurueck
         switch (action) {
            case 1: textarea.appendText("bal(rechter Sohn) = +1 oder 0"+"\n");
                    textarea.appendText("=> Links-Rotation notwendig"+"\n");
                    rot1Tree = new VerkListe();    //leftrotation
                    rot1Tree.Split(allMoves);
                    arrangeMove(initTree,rot1Tree);
                    oldTree = rot1Tree;
                    //initTree = rot1Tree;
                    //rot1Tree = null;
                    break;
            case 2: textarea.appendText("bal(linker Sohn) = -1 oder 0"+"\n");
                    textarea.appendText("=> Rechts-Rotation notwendig"+"\n");
                    rot1Tree = new VerkListe();    //rightrotation
                    rot1Tree.Split(allMoves);
                    arrangeMove(initTree,rot1Tree);
                    oldTree = rot1Tree;
                    //initTree = rot1Tree;
                    //rot1Tree = null;
                    break;
            case 3: textarea.appendText("bal(rechter Sohn) = -1"+"\n");
                    textarea.appendText("=> Links/Rechts-Rotation notwendig"+"\n");
                    rot1Tree = new VerkListe();         //left-right
                    rot2Tree = new VerkListe();
                    rot1Tree.Split(allMoves);
                    rot2Tree.Split(allMoves);
                    arrangeMove(initTree,rot1Tree);
                    arrangeMove(rot1Tree,rot2Tree);
                    oldTree = rot2Tree;
                    //initTree = rot2Tree;
                    //rot1Tree = null;
                    //rot2Tree = null;
                    break;
            case 4: textarea.appendText("bal(linker Sohn) = +1"+"\n");
                    textarea.appendText("=> Rechts/Links-Rotation notwendig"+"\n");
                    rot1Tree = new VerkListe();          //right-left
                    rot2Tree = new VerkListe();
                    rot1Tree.Split(allMoves);
                    rot2Tree.Split(allMoves);
                    arrangeMove(initTree,rot1Tree);
                    arrangeMove(rot1Tree,rot2Tree);
                    oldTree = rot2Tree;
                    //initTree = rot2Tree;
                    //rot1Tree = null;
                    //rot2Tree = null;
                    break;
         }
         rot1Tree = null;
         rot2Tree = null;
         initTree = null;
         if(allMoves.Length()>0) {
            initTree = new VerkListe();
            allMoves.GoFirst();
            action = allMoves.GetBalance();
            allMoves.GoFirst();
            initTree.Split(allMoves);            
         }
         if(allMoves.Length()==0) puCl.posOld = initTree.Clone(); //letzte Baumstellung merken fr naechste Aktion 
      }            
   }
   /**Zeichnet AVL-Baum. Entweder mit markierten Suchpfad oder mit blauem
   Pfeil(zeigt Rotationsrichtung an), falls gerade Rotation stattfindet.
   Bewegungsdaten in stack, Beginn bei Element b in stack, analog Bewegungsende e,
   AVL-Baumdaten tree(im Hintergrund bei Rotation),Knotenfarbe newNode,
   Balancefaktorfarbe newBal.*/
   public void printNodeWithData2(VerkListe stack,VerkListe lines,int b,int e,
                                  VerkListe tree,Color newNode,Color newBal){
      clear(offg);
      printTree(offg,tree,newNode,newBal);
      if (stack != null && stack.Length()>0) {
         offg.setFont(dataFont);
         offg.setColor(Color.blue);
         for(int i = b;i<=e; i++) {
            stack.Go(i);
            int x = stack.Getx1();
            int y = new Float(stack.Gety2()).intValue();
            int data = stack.GetData();
            if (data>=0) { //nur bei Rotationen Nodes zeichenen
               offg.drawOval(x,y,30,30);
               int shift = 8;
               if (data < 10) shift = 13;
                  else if ( data < 100) shift = 10;
               offg.drawString(""+data,x+shift,y+23);
               offg.drawLine(x+4,y+11,x+26,y+11);
               offg.drawArc(x+3,y,30,30,90,-180);
            }
            if(b==e) scroll.setScrollbarValues(x,y);
         }
         if (forward && !synchron) {  //print thick lines if not synchron
            for(int k = 1;k<=e; k++) {
               stack.Go(k);
               printThickLines(stack,true);
            }
         }
         if (synchron) drawArrow();

      }
      scroll.sentImage(offscreen);
   }

   /**Fuehrt eine Rotation im Animationsfenster durch. Berechnet
   Bewegungsdaten,zeichnet blauen Pfeil und zeichnet Bewegungen.*/
   public void arrangeMove(VerkListe stack1,VerkListe stack2) {
      moveNode1 = computeMoveNode(stack1,stack2);
      if (moveNode1 != null) {
         int beginTmp = begin;int endTmp = end;boolean synchronTmp=synchron;
         begin = 1;end=moveNode1.Length();
         synchron = true;
         findRotCoord(moveNode1);
         while (! moveReady(moveNode1)) {
            try {
               printNodeWithData2(moveNode1,begin,end,stack1,Color.gray,Color.gray);
               nextCoordinate(moveNode1,begin,end);
               thread.sleep(sleepTime);
               stopOrRun();
            }
            catch (InterruptedException e) {}
         }
         clear (offg);
         printTree(offg,stack2,Color.black,Color.red);
         scroll.sentImage(offscreen);
         begin=beginTmp;end=endTmp;synchron=synchronTmp;
      }
   }
   /**Fuehrt eine Vertauschungsbewegung durch. Falls ein zu loeschender
   Schluessel im AVL-Baum ein innerer ist, muss mit groesstem Schluessel
   im linken Teilbaum vertauscht werden. Diese Vorgang wird hier optisch
   dargestellt.*/
   public void arrangeSwapMove(VerkListe stack1,VerkListe stack2,VerkListe stack3) {
      moveNode2 = computeMoveNode(stack1,stack2);
      if (moveNode2 != null) {
         int beginTmp = begin;int endTmp = end;boolean synchronTmp = synchron;
         begin = 1;end=moveNode2.Length();
         synchron = true;
         //findRotCoord(moveNode1);
         while (! moveReady(moveNode2)) {
            try {
               printNodeWithData2(moveNode2,begin,end,stack3,Color.gray,Color.gray);
               nextCoordinate(moveNode2,begin,end);
               thread.sleep(sleepTime);
               stopOrRun();
            }
            catch (InterruptedException e) {}
         }
         begin = beginTmp; end = endTmp;synchron = synchronTmp;
      }
   }
   /**Veranschaulicht optisch den Suchvorgang des zu loeschenden Schluessel.
   Veranlasst und zeigt an, welche Schritte beim Suchen notwendig sind.*/
   public void suchVorgang() {
      if (begin > einfPfad.Length()) einfPfad.Go(1);
         else einfPfad.Go(begin);
      int data = einfPfad.GetData();
      int bal = einfPfad.GetBalance();
      int dataBlatt = -1;
      VerkListe stack1,stack2;
      switch (bal) {
         case 0 : if(!innererKnoten) {
                     if(del > data)
                        textarea.appendText(del+" > " + data + " => Suche weiter im rechten Teilbaum"+"\n");
                     else
                        textarea.appendText(del+" < " + data + " => Suche weiter im linken Teilbaum"+"\n");
                  }
                  break;
         case -1: textarea.appendText("Schluessel " +del+" gefunden. Ist innerer Knoten."+"\n");
                  textarea.appendText("=> Suche groessten Schluessel im linken Teilbaum."+"\n");
                  textarea.appendText("   Dieser Knoten kann kein innerer Knoten sein."+"\n");
                  textarea.appendText("   Vertausche dann diesen Schluessel mit "+del+"."+"\n");
                  innererKnoten = true;
                  break;
         case -2: case -3:
                  //einfPfad.Go(begin+1);
                  //dataBlatt = einfPfad.GetData();
                  if (!innererKnoten) {
                     textarea.appendText("Schluessel " +del+" gefunden"+"\n");
                  }
                  else {
                     textarea.appendText("Groesster Schluessel im linken Teilbaum gefunden = "+data+"\n");
                     textarea.appendText("Vertausche "+del+" und "+data+"\n");
                     stack1 = oldTree.prepareSwap(del,data);
                     stack2 = oldTree.prepareSwap(data,del);
                     arrangeSwapMove(stack1,stack2,oldTree);
                     oldTree.swapData(del,data);
                     einfPfad.swapData(del,data);
                     printNodeWithData2(moveInsert,null,begin,end,oldTree,Color.black,Color.red);
                     wait(1000);
                     innererKnoten = true;
                  }
                  if (bal==-2) {
                     textarea.appendText("Zu loeschender Knoten hat noch linkes Blatt"+"\n");
                     dataBlatt = einfPfad.GetLeft();
                  }
                  else {
                     textarea.appendText("Zu loeschender Knoten hat noch rechtes Blatt"+"\n");
                     dataBlatt = einfPfad.GetRight();
                  }
                  textarea.appendText("Vertausche nun Schluessel "+dataBlatt+" und "+del+"\n");
                  textarea.appendText("und loesche Blatt "+del+"\n");
                  stack1 = oldTree.prepareSwap(del,dataBlatt);
                  stack2 = oldTree.prepareSwap(dataBlatt,del);
                  arrangeSwapMove(stack1,stack2,oldTree);
                  oldTree.swapData(del,dataBlatt);                  
                  //printNodeWithData2(moveInsert,null,begin,end,oldTree,Color.black,Color.red);
                  //System.out.println("wait");
                  //wait(1000);
                  oldTree.delete(del);
                  einfPfad.delete(del);
                  oldTree.search(dataBlatt);
                  oldTree.SetLeft(0);
                  oldTree.SetRight(0);
                  oldTree.SetBalance(0);
                  printNodeWithData2(moveInsert,null,begin,end,oldTree,Color.black,Color.red);
                  wait(1000);
                  break;
         case -4: if (!innererKnoten) {
                     textarea.appendText("Schluessel "+del+" gefunden"+"\n");
                  }
                  else {
                     textarea.appendText("Groesstes Element im linken Teilbaum gefunden = "+data+"\n");
                     textarea.appendText("Vertausche "+del+" und "+data+"\n");
                     stack1 = oldTree.prepareSwap(del,data);
                     stack2 = oldTree.prepareSwap(data,del);
                     arrangeSwapMove(stack1,stack2,oldTree);
                     oldTree.swapData(data,del);
                     einfPfad.swapData(data,del);
                     //printNodeWithData2(moveInsert,null,begin,end,oldTree,Color.black,Color.red);
                     wait(1000);
                  }
                  textarea.appendText("Zu loeschender Knoten ist ein Blatt"+"\n");
                  textarea.appendText("Loesche Blatt "+del+"\n");
                  einfPfad.delete(del);
                  oldTree.delete(del);               //Element im Baum loeschen
                  if (oldTree.Length()==0) {    //Baum ist nun leer
                    puCl.posOld = null;         //thread abbrechen
                    clear (offg);
                    scroll.sentImage(offscreen);
                    puCl.einfuegenOk.enable();
                    thread.stop();
                  }
                  einfPfad.GoLast();                 //und beim Vorgaenger
                  int vorgaenger = einfPfad.GetData();
                  initTree.search(vorgaenger);
                  oldTree.search(vorgaenger); //die Hoehe korrigieren
                  if (initTree.GetLeft()==0) oldTree.SetLeft(0);
                  if (initTree.GetRight()==0) oldTree.SetRight(0);
                  innererKnoten = false;
                  break;
         default: textarea.appendText("bad input");
                  break;
      }
   }
   /**Uepberprueft bei jedem Knoten beim Zurueckgehen des Suchpfades, ob eine Rotation notwendig ist.
   Bei Erreichen jedes Knoten wird Ergebnis der Untersuchung im
   Textfenster angezeigt.*/
   public void correctBalance() {
      einfPfad.GoLast();
      //System.out.println("d"+einfPfad.GetData());
      int data = einfPfad.GetData();
      einfPfad.delete(einfPfad.GetData());
      initTree.search(data);
      int left = initTree.GetLeft();
      int right = initTree.GetRight();
      int newBalance = right - left;
      textarea.appendText("Untersuche Knoten "+data+"\n");
      textarea.appendText("Hoehe rechter Teilbaum = "+right+"\n");
      textarea.appendText("Hoehe linker Teilbaum  = "+left+"\n");
      textarea.appendText("Neuer Balancefaktor = "+newBalance+"\n");
      if(Math.abs(newBalance)==2) {
         textarea.appendText("=> Rotation notwendig"+"\n");
         rotation(action);
      } else {
         textarea.appendText("=> keine Rotation notwendig"+"\n");
         oldTree.search(data);
         oldTree.SetBalance(newBalance);
      }
   }

   /**Suchbewegung beim Loeschen berechnen und stack zurckgeben.
   Bewgungs-Koordinaten von Wurzel bis zum loeschenden Schluessel werden ermittelt.*/
   public VerkListe computeMoveDelete(VerkListe stack) {
      VerkListe tmp = null;
      if (stack != null) {
         tmp = new VerkListe();
         //float yfloat = -30; float mfloat =5;
         //tmp.PushToEnd(-1,80,1300,yfloat,mfloat,0);
         int laenge = stack.Length();
         for (int i=1;i<laenge;i++) {
            stack.Go(i);
            Integer y1Int = new Integer(stack.Gety1());
            float y1float = y1Int.floatValue() + 25;
            int startx = 0;
            int x1int = stack.Getx1();
            Integer x1Int = new Integer(x1int);
            float x1float = x1Int.floatValue();
            stack.Go(i+1);
            Integer y2Int = new Integer(stack.Gety1());
            float y2float = y2Int.floatValue();
            int x2int = stack.Getx1();
            Integer x2Int = new Integer(x2int);
            float x2float = x2Int.floatValue();
            int steps = 0;
            int steplength = 0;
            int direction = 0;
            float m = 0;
            if (x1int>x2int) {
               x1float = x1float + 4 -15;
               m = (y2float-y1float) / (x2float-x1float);
               m=0-m;
               //tmp.PushToEnd(-1,5,x1int,y1float,0,-5);
               startx = x1int;
               x1int = x1int + 4 -15;
               steps = Math.abs(x1int-x2int);
               steplength = steps / 5;
               direction = 0-steplength;
            }
            else {
               x1float = x1float + 26 -15;
               m = (y2float-y1float) / (x2float-x1float);
               //tmp.PushToEnd(-1,5,x1int,y1float,0,5);
               startx = x1int;
               x1int = x1int + 26 - 15;
               steps = Math.abs(x1int-x2int);
               steplength = steps / 5;
               direction = steplength;
            }
            tmp.PushToEnd(-1,steps,x1int,y1float,m,direction);
            tmp.SetLeft(x1int);
            tmp.Sety1(new Float(y1float).intValue());
            //if (laenge>(i+1)) {
            //   m=5;
            //   tmp.PushToEnd(data,25,x2int,y2float,m,0);
            //}
         }
      }
      return tmp;
   }
   /**Thread sleep fuer time 1/1000 Sekunden*/
   public void wait(int time) {
      try {
         thread.sleep(time);
      }
      catch (InterruptedException e) {}
   }
}
/**Dient dazu die Animationsgeschwindigkeit zu regeln. Dazu wird
ein neues Fenster geoffnet mit einer Scrollbar, in der die Geschwindigkeit
gerregelt werden kann.*/
class SpeedSelect extends Frame {
   private Label currentSpeed;
   private int sleepTime;
   private Scrollbar speed;
   private Button ok,cancel;
   private Font font = new Font("Dialog",Font.BOLD,15);
   private MainFrame puCl;
   /**Konstruktur. Baut neues Fenster auf mit Scrollbar zur Wahl der Geschwindigkeit*/
   public SpeedSelect(MainFrame publicClass) {
      puCl = publicClass;
      sleepTime = puCl.sleepTime;
      setTitle("Animationsgeschwindigkeit regulieren");
      Panel label = new Panel();
      //p.setLayout(new GridLayout(1,2));
      label.setLayout(new BorderLayout());
      label.add("Center",currentSpeed = new Label("Pause zwischen Bewegungen (Tsd.Sec) : "+sleepTime));
      Panel scrollbar = new Panel();
      scrollbar.setLayout(new BorderLayout());
      scrollbar.add("Center",speed = new Scrollbar(Scrollbar.HORIZONTAL,0,0,10,500));
      speed.setPageIncrement(20);
      speed.setLineIncrement(5);
      speed.setValue(sleepTime);
      Panel button = new Panel();
      button.setLayout(new FlowLayout());
      button.add(ok = new Button("Ok"));
      button.add(cancel = new Button ("Cancel"));
      this.setLayout(new BorderLayout());
      this.setFont(font);
      add("North",label);
      add("Center",scrollbar);
      add("South",button);
   }
   public boolean handleEvent(Event evt) {
      if (evt.id == Event.WINDOW_DESTROY) {
         this.hide();
         this.dispose();
      }
      else if (   evt.id ==Event.SCROLL_ABSOLUTE
               || evt.id ==Event.SCROLL_LINE_DOWN
               || evt.id ==Event.SCROLL_LINE_UP
               || evt.id ==Event.SCROLL_PAGE_DOWN
               || evt.id ==Event.SCROLL_PAGE_UP) {
         currentSpeed.setText("Pause zwischen Bewegungen (Tsd.Sec.) : "+speed.getValue()+"   ");
         return true;
      }
      else if(evt.target == ok) {
         puCl.sleepTime = speed.getValue();
         this.hide();
         this.dispose();
         return true;
      }
      else if(evt.target == cancel) {
         this.hide();
         this.dispose();
         return true;
      }      
      return super.handleEvent(evt);
   }
}
/**Animationsfenster mit Scrollbars in dem die Animation ablaeuft.*/
class ScrollableCanvas extends Panel {
   private Image offscreenImage = null;
   public Scrollbar hbar,vbar;
   Canvas canvas;
   /**Breite des sichtbaren Bereichs des Animationsfenster = Breite Panel - Breite Scrollbar.*/
   private int canvas_width;
   /**Hoehe des sichtbaren Breich des Animationsfenster = Breite Panel - Breite Scrollbar.*/
   private int canvas_height;
   /**Aktuelle Position der vertikalen Scrollbar.*/
   public int offX = -1;
   /**Aktuelle Position der horizontalen Scrollbar.*/
   public int offY = -1;

   /**Legt ein Canvas und zwei Scrollbars an und legt diese in ein Panel ab.
   Benutzt BorderLayout, um die Scrollbars horizontal und vertikal anzulegen.
   Wenn die Scrollbars benutzt werden, verschiebt sich das Animationsfenster,
   in dem die Animation ablaeuft (Offscreen).*/
   public ScrollableCanvas(){
      canvas = new Canvas();
      hbar = new Scrollbar(Scrollbar.HORIZONTAL);
      vbar = new Scrollbar(Scrollbar.VERTICAL);
      hbar.setLineIncrement(10);
      vbar.setLineIncrement(10);
      this.setLayout(new BorderLayout(0,0));
      //this.add("Center",canvas);
      this.add("South",hbar);
      this.add("East",vbar);
   }
   /**Schnittstelle, um Klasse ScrollableCanvas die Offscreen, in der alle Zeichnungen aufgebaut werden, zu senden.
   Nachdem erhalt einer neuen Offscreen, wird Offscreen unmittelbar am
   Bildschirm ausgegeben.*/
   public void sentImage(Image newoffscreenImage) {
      offscreenImage=newoffscreenImage;
      repaint();
      //this.update(canvas.getGraphics());
   }
   public void update (Graphics g) {
      paint(g);
   }

   /**Gibt Inhalt der Offscreen aus.
   Die offset_x und offset_y Variablen spezifizieren welcher Teil von
   der grosseren Offscreen im relativeren kleineren Canvas gezeigt werden soll.
   Durch Bewegen der Scrollbars veraendern sich offset_x und offset_y
   und die veranlasst das Scrollen. Nur der Teil, der Offscreen der sich
   im Canvas sich befindet ist sichtbar, der Rest wird abgeschnitten.*/
   public void paint (Graphics g) {
      if (offscreenImage != null) {
        checkScrollbar(vbar);
        checkScrollbar(hbar);
        g.drawImage(offscreenImage,-hbar.getValue(),-vbar.getValue(),null);
      }
   }

   /**Ermittelt aktuelle Koordinaten der Scrollbars.*/
   public boolean mouseDown(Event event,int x, int y) {   
      offX = x + hbar.getValue();
      offY = y + vbar.getValue();
      //return super.handleEvent(event);
      return false;
   }
   /**Ermittelt aktuelle Koordinaten der Scrollbars.*/
   public boolean mouseMove(Event event, int x, int y) {
      offX = x + hbar.getValue();
      offY = y + vbar.getValue();
      return false;
   }
   /**Behandelt die Scrollbar-Events. Aktualisiert die offset_x und
   offset_y - Variablen. Nach jedem des Events wird ein repaint()
   durchgefuehrt.*/
   public boolean handleEvent(Event e) {
      if ( e.target==hbar ) {
         switch(e.id) {
            case Event.SCROLL_LINE_UP:
               hbar.setValue(hbar.getValue() - hbar.getLineIncrement());break;
            case Event.SCROLL_LINE_DOWN:
               hbar.setValue(hbar.getValue() + hbar.getLineIncrement());break;
            case Event.SCROLL_PAGE_UP:
               hbar.setValue(hbar.getValue() - hbar.getPageIncrement());break;
            case Event.SCROLL_PAGE_DOWN:
               hbar.setValue(hbar.getValue() + hbar.getPageIncrement());break;
            case Event.SCROLL_ABSOLUTE:
               hbar.setValue(((Integer)e.arg).intValue());break;
         }
         //this.update(canvas.getGraphics());
         repaint();
         return true;
      }
      else if(e.target==vbar) {
         switch(e.id){
            case Event.SCROLL_LINE_UP:
               vbar.setValue(vbar.getValue()-vbar.getLineIncrement());break;
            case Event.SCROLL_LINE_DOWN:
               vbar.setValue(vbar.getValue()+vbar.getLineIncrement());break;
            case Event.SCROLL_PAGE_UP:
               vbar.setValue(vbar.getValue()-vbar.getPageIncrement());break;
            case Event.SCROLL_PAGE_DOWN:
               vbar.setValue(vbar.getValue()+vbar.getPageIncrement());break;
            case Event.SCROLL_ABSOLUTE:
               vbar.setValue(((Integer)e.arg).intValue());break;
         }
         //this.update(canvas.getGraphics());
         repaint();
         return true;
      }
      //if we did not handle it above, pass it on to the superclass
      //handleEvent routine, which will check its type and call
      //the mouseDown(),mouseDrags(), and other methods.
      return super.handleEvent(e);
   }

   /**Diese Methode wird aufgerufen, wenn sich die Groesse des Fensters sich aendert.
   Dadurch koennen die Scrollbar-Werte und Canvasdimension
   aktualisiert werten.*/
   public synchronized void reshape(int x, int y, int width, int height) {
      //do the real stuff
      super.reshape(x,y,width,height);
      //Update our scrollbar page size;
      Dimension hbar_size = hbar.size();
      Dimension vbar_size = vbar.size();
      canvas_width = width - vbar_size.width;
      canvas_height = height - hbar_size.height;
      //hbar.setValues(hbar.getValue(),canvas_width,0,offscreenImage.getWidth(this));
      //vbar.setValues(vbar.getValue(),canvas_height,0,offscreenImage.getHeight(this));
      hbar.setValues(hbar.getValue(),canvas_width,0,2600);
      vbar.setValues(vbar.getValue(),canvas_height,0,400);
      hbar.setPageIncrement(canvas_width/2);
      vbar.setPageIncrement(canvas_height/2);
      //printDaten();
      repaint();
   }   
   /**Bewegt die Scrollbars, damit man alle Animationaktionen sehen kann.
   Wenn bei einem Bewegungsvorgang die Aktion sich aus dem sichtbaren
   Bereich entfernt werden Scrollbarwerte angepasst.*/
   public void setScrollbarValues(int x, int y) {
      int hbarValue = hbar.getValue();
      int vbarValue = vbar.getValue();
      int hbarVisible = hbar.getVisible();
      int vbarVisible = vbar.getVisible();
      int borderVisible = 40;
      int delta = y + 45 - (vbarValue + vbarVisible);
      if (delta > 0) vbar.setValue(vbarValue+delta);
      else {
         delta = vbarValue - (y - 60);
         if(delta > 0) vbar.setValue(vbarValue-delta);
      }
      delta = x + 40 - (hbarValue + hbarVisible);
      if (delta > 0) hbar.setValue(hbarValue+delta);
      else {
         delta = hbarValue - x;
         if(delta > 0) hbar.setValue(hbarValue-delta);
      }
   }
   /**Ueberprueft, ob Scrollbarwert sich im erlaubten Bereich befindet.*/
   public void checkScrollbar(Scrollbar tmp) {
      int value = tmp.getValue();
      int valueMax = tmp.getMaximum() - tmp.getVisible();
      if(value>valueMax) tmp.setValue(valueMax);
         else if(value < 0) tmp.setValue(0);
   }
   /**Uebergibt offset_x*/
   public int getX() {return offX;}
   /**Uebergibt offset_y*/
   public int getY() {return offY;}
}
/**Simuliert das Loeschen und Einfuegen von Schluesseln. Einige
werden durch Programmanweisungen zuerst eingefuegt und geloescht.
Dient dazu dem Anwender die Animation vorzufuehren. Dabei werden 
interessante Faelle vorgefuehrt.*/
class Demo extends Object implements Runnable {
  Thread thread=null;
  private MainApplet mainApplet;
  MainFrame puCl;
  public Demo(MainFrame newPC,MainApplet mApplet) {
    puCl = newPC;
    mainApplet = mApplet;
  }
   public void start() {
      if(thread==null) {
         thread = new Thread(mainApplet.threadGroup,this);
         thread.start();
      }
   }
   public void stop() {
      if(thread!=null && thread.isAlive()) thread.stop();
   }

   public void run() {
     int[] feld;
     feld = new int[10];
     feld[0]=400;
     feld[1]=500;
     feld[2]=600;
     feld[3]=200;
     feld[4]=300;
     feld[5]=700;
     feld[6]=100;
     feld[7]=400;

     try {
       for (int i =0;i<=6;i++) {
         puCl.textfield.setText(""+feld[i]+ " einfuegen - Demo");
         puCl.addNewElement(feld[i],thread);
         thread.suspend();
         //puCl.textfield.setText("");
       }
     }
     catch (NumberFormatException nfe) {
       puCl.textarea.appendText("Das ist keine Zahl"+"\n");
     }
     try {
       puCl.textfield.setText(""+feld[7]+ " loeschen - Demo");
       puCl.deleteElement(feld[7],thread);
       thread.suspend();
       puCl.textfield.setText("");
     }
     catch (NumberFormatException nfe) {
        puCl.textarea.appendText("Das ist keine Zahl"+"\n");
     } 
   }
}   
/**Ist das Hilfemenu der AVL-Baum-Animation. In einem seperatem Fenster
werden alle Menuoptionen erklaert.*/
class Hilfe extends Frame {
    private Choice auswahl;
    private TextArea text=new TextArea();
    /**Konstruktor: Legt neues Fenster mit Auswahl und Textarea an.*/
    public Hilfe(String topic) {
        super("Hilfe zur interaktiven AVL-Baum-Animation");
        auswahl=new Choice();
        auswahl.addItem("GESAMT");
        auswahl.addItem("Kurze Einleitung");
        auswahl.addItem("Schluessel einfuegen");
        auswahl.addItem("Schluessel loeschen");
        auswahl.addItem("Clear");
        auswahl.addItem("Ungeschehen");
        auswahl.addItem("Animationsgeschwindigkeit");
        auswahl.addItem("AVL-Baum/Neu");
        auswahl.addItem("AVL-Baum/Demo starten");
        auswahl.addItem("Stop");
        auswahl.addItem("Run");
        auswahl.addItem("Next Step");
        auswahl.select(topic);
        if (topic.equals("Hilfe")) {
          text.setText(allgemeineinfos()+einfuegen()+loeschen()+clear()+
          stop()+run()+nextstep()+ungeschehen()+animationsgeschwindigkeit()+
          avlbaumneu()+demo());
        }
        else if (topic.equals("Kurze Einleitung")) {
          text.setText(allgemeineinfos());
        }
        else if (topic.equals("Schluessel einfuegen")) {
          text.setText(einfuegen());
        }
        else if (topic.equals("Schluessel loeschen")) {
          text.setText(loeschen());
        }
        else if (topic.equals("Clear")) {
          text.setText(clear());
        }
        else if (topic.equals("Ungeschehen ")) {
          text.setText(ungeschehen());
        }
        else if (topic.equals("Geschwindigkeit")) {
          text.setText(animationsgeschwindigkeit());
        }
        else if (topic.equals("AVL-Baum/Neu")) {
          text.setText(avlbaumneu());
        }
        else if (topic.equals("AVL-Baum/Demo starten")) {
          text.setText(demo());
        }
        else if (topic.equals("Stop")) {
          text.setText(stop());
        }
        else if (topic.equals("Run")) {
          text.setText(run());
        }
        else if (topic.equals("Next Step")) {
          text.setText(nextstep());
        }
        text.setEditable(false);
        add("North", auswahl);
        add("Center",text);
        add("South", new Button ("Schliessen"));
        pack();
        resize(500,300);
        move(200,150);
        show();
    }
    private String allgemeineinfos () {
       return "Kurze Einleitung :\n"+
       "Im Informationfeld links oben werden Erklaerungen zu\n"+
       "den Vorgaengen Einfuegen und Loeschen ausgegeben.\n"+
       "Dieses wird nach jedem Vorgang neu aufgebaut.\n"+
       "Im Eingabefeld rechts oben koennen nummerische Eingaben\n"+
       "von 0 bis 999 gemacht werden und die Operationen\n"+
       "Einfuegen/Loeschen von Knoten im AVL-Baum durchgefuehrt werden.\n"+
       "Zur Erleichterung des Benutzers gibt es noch folgende Hilfen:\n"+
       "Einfach-Klick ueber Knoten : Potentielle Eingabe-Schluessel\n"+
       "links und rechts von Knoten werden aufgelistet.\n"+
       "Doppel-Klick ueber Knoten : Angeglickter Schluessel wird im\n"+
       "angezeigt und muss mit Button Loeschen nur noch bestaetigt werden.\n"+
       "Animation verfolgen : Durch Betaetigen des Button Stop wird\n"+
       "die Animation angehalten. Analog Run. Mit Druecken Next Step\n"+
       "kann die Animation schrittweise verfolgt werden.\n\n";
    }
    private String ungeschehen () {
        return "Ungeschehen: \n"+
        "Alle bisher erfolgte Schritte koennen\n"+
        "sukzessive ungeschehen gemacht werden.\n\n";
    }

    private String animationsgeschwindigkeit() {
        return "Animationsgeschwindigkeit :\n"+
        "Die Geschwindigkeit der Animation ist regelbar.\n"+
        "Dabei muss die Pause zwischen den einzelnen\n"+
        "Animation-Bewegungen bestimmt werden, damit eine\n"+
        "Anpassung an die jeweilige Rechnerleisung moeglich ist.\n\n";
    }

    private String avlbaumneu () {
        return "AVL-Baum/Neu :\n"+
        "Aktueller AVL-Baum wird geloescht.\n\n";
    }
    private String demo () {
        return "AVL-Baum/Demo starten :\n"+
        "Zur Veranschaulichung des Anwenders werden einige\n"+
        "interresante Einfuege/Loesch-Operationen demonstriert.\n\n";
    }
    private String einfuegen () {
        return "Schluessel einfuegen :\n"+
        "Durch diesen Button koennen neue Knoten in den Baum\n"+
        "eingefuegt werden.\n"+
        "Zuvor muss eine Eingabe im Eingabefenster erfolgt sein.\n"+
        "Eingabe durch Nummer-Buttons oder Tastatur.\n\n"+
        "Bestaetigen mit Button Schluessel einfuegen oder Enter-Taste.\n"+
        "Durch Klicken auf einen existierenden Knoten im Animation-\n"+
        "Fenster kann der Anwender herausfinden, welche Schluessel\n"+
        "links und rechts von diesem Knoten noch eingefuegt werden\n"+
        "koennen.\n\n";
    }
    private String loeschen () {
        return "Schluessel loeschen :\n"+
        "Durch diesen Button kann ein existierender Knoten im\n"+
        "im AVL-Baum geloescht werden.\n"+
        "Zuvor muss eine Eingabe im Eingabefenster erfolgt sein.\n"+
        "Eingabe durch Nummer-Buttons oder Tastatur.\n\n"+
        "Bestaetigen mit Button Schluessel loeschen oder Entf-Taste.\n\n"+
        "Oder durch Anklicken auf den Knoten im bestehenden\n"+
        "AVL-Baum der geloescht werden soll, erscheint der\n"+
        "jeweilige Schluessel im Eingabe-Fenster.\n\n";
    }

    private String clear() {
        return "Clear :\n"+
        "Eingabefeld wird geloescht.\n\n";
    }

    private String stop() {
        return "Stop :\n"+
        "Erscheint nur, wenn Animation aktiv ist.\n"+
        "Button bewirkt, dass Animation angehalten wird.\n\n";
    }

    private String run() {
        return "Run :\n"+
        "Erscheint nur, wenn Animation angehalten wurde.\n"+
        "Durch Anklicken wird Animation wieder fortgefuehrt\n\n";
    }

    private String nextstep () {
        return "Next Step :\n"+
        "Wenn die Animation genauer verfolgt werden soll,\n"+
        "ist dieser Button von Hilfe.\n"+
        "Entweder wird die laufende Animation angehalten\n"+
        "und naechster Animationschritt gezeigt,\n"+
        "oder bei einer unterbrochenen Animation laeuft diese\n"+
        "nur bis zum naechsten Schritt.\n\n";
    }
    public boolean action (Event e, Object o) {
        if (e.target instanceof Button && "Schliessen".equals(o)) {
            dispose();
        }
        else if(e.id == Event.WINDOW_DESTROY) {
            dispose();
        }
        else if (e.target.equals(auswahl)) {
            if (o.equals("GESAMT")) {
               text.setText(allgemeineinfos()+einfuegen()+loeschen()+clear()+
               stop()+run()+nextstep()+ungeschehen()+
                avlbaumneu()+demo());
            }
            else if (o.equals("Kurze Einleitung")) {
                text.setText(allgemeineinfos());
            }
            else if (o.equals("Schluessel einfuegen")) {
                text.setText(einfuegen());
            }
            else if (o.equals("Schluessel loeschen")) {
                text.setText(loeschen());
            }
            else if (o.equals("Clear")) {
                text.setText(clear());
            }
            else if (o.equals("Stop")) {
                text.setText(stop());
            }
            else if (o.equals("Run")) {
                text.setText(run());
            }
            else if (o.equals("Next Step")) {
                text.setText(nextstep());
            }
            else if (o.equals("Ungeschehen")) {
                text.setText(ungeschehen());
            }
            else if (o.equals("Animationsgeschwindigkeit")) {
                text.setText(animationsgeschwindigkeit());
            }
            else if (o.equals("AVL-Baum/Neu")) {
                text.setText(avlbaumneu());
            }
            else if (o.equals("AVL-Baum/Demo starten")) {
                text.setText(demo());
            }
        }
        return true;
    }
}
/**Hauptfenster mit Textarea, Kontrollpanel, Animationsfenster und Steuerungsbuttons.
Das zentrale Event-Handling wird von hier aus gesteuert, d.h. alle
Klassenaufrufe (Ausnahme MainApplet) erfolgen aus dieser Klasse.*/
class MainFrame extends Frame {
    private Button[] figure; //figures from 0 to 9
    //Buttons fuer das Panel Control(Einfuegen und Loeschen)
    public Button einfuegenOk,loeschenOk,clearTextfield; 
    public Button run,stop,nextStep; //buttons to control the animation
    public TextArea textarea;        // A text window
    public TextField textfield;
    public ScrollableCanvas scribble; // An area to draw in
    private Panel controlPanel,stopRunPanel;
    public Image offscreen = null;
    public int sleepTime = 20;
    private int imageWidth = 2600;
    private int imageHeight = 450;
    public AVLTree Tree = new AVLTree();
    public VerkListe posOld = null;
    public VerkListe undo = null;
    public Stack history = null; //merkt sich alle Aktionen
    private MainFrame publicClass;
    private Font font = new Font("Dialog",Font.BOLD,11);
    private int focusKey = -1; //-1=keine sonst jeweiliger Key hervorgehoben
    private MenuItem ungeschehen,baum_neu;
    private Panel northPanel;
    public Graphics offg;
    private Thread thread=null;
    private MainApplet mainApplet;
    /**Baut ein neues Fenster mit Menueleitse,Textarea, Kontrollpanel, Animationsfenster und Steuerungsbuttons auf.*/
    public MainFrame(Image newoffscreen,MainApplet m) {
      super("AVL-Baum");
      //resize(1000,700);
      //move(5,5);
      setFont(font);
      MenuBar menubar = new MenuBar();
      Menu fenster = new Menu("Fenster");
      fenster.add(new MenuItem("Schliessen"));
      Menu bearbeiten = new Menu("Bearbeiten");
      ungeschehen = new MenuItem("Ungeschehen");
      bearbeiten.add(ungeschehen);
      bearbeiten.add(new MenuItem("Geschwindigkeit..."));
      ungeschehen.disable();
      Menu baum = new Menu("AVL-Baum");
      baum.add("Demo starten");
      baum_neu = new MenuItem("Neu");
      baum.add(baum_neu);
      baum_neu.disable();
      bearbeiten.add(baum);
      Menu hilfe = new Menu(" ? ");
      MenuItem h = new MenuItem("Hilfe");
      hilfe.add(h);
      hilfe.addSeparator();
      //hilfe.add(new MenuItem("GESAMT"));
      hilfe.add(new MenuItem("Kurze Einleitung"));
      hilfe.add(new MenuItem("Schluessel einfuegen"));
      hilfe.add(new MenuItem("Schluessel loeschen"));
      hilfe.add(new MenuItem("Clear"));
      hilfe.add(new MenuItem("Ungeschehen "));
      hilfe.add(new MenuItem("Geschwindigkeit"));
      hilfe.add(new MenuItem("AVL-Baum/Neu"));
      hilfe.add(new MenuItem("AVL-Baum/Demo starten"));
      hilfe.add(new MenuItem("Stop"));
      hilfe.add(new MenuItem("Run"));
      hilfe.add(new MenuItem("Next Step"));
      menubar.add(fenster);
      menubar.add(bearbeiten);
      menubar.add(hilfe);
      setMenuBar(menubar);

      publicClass = this;
      offscreen = newoffscreen;
      mainApplet = m;
      history = new Stack();

      //Create a scrolling canvas to draw in
      scribble = new ScrollableCanvas();
      scribble.sentImage(offscreen);

      //Create pushbuttons
      nextStep = new Button(" Next Step ");
      stop = new Button(" Stop ");      
      run = new Button(" Run ");
      run.disable();
      nextStep.disable();
      einfuegenOk = new Button("Schluessel einfuegen"); // insert key
      loeschenOk = new Button("Schluessel loeschen"); //delete key
      clearTextfield = new Button("Clear"); //clears textfield of control panel
      figure = new Button[10];
      figure[0] = new Button("0");figure[1] = new Button("1");
      figure[2] = new Button("2");figure[3] = new Button("3");
      figure[4] = new Button("4");figure[5] = new Button("5");
      figure[6] = new Button("6");figure[7] = new Button("7");
      figure[8] = new Button("8");figure[9] = new Button("9");

      //Create a one-line text field, and multiline text area.
      textarea = new TextArea(8,40);
      textarea.setEditable(false);
      textfield = new TextField("",3);
      //textfield.setEditable(false);

      //Panel for buttons of the controlPanel
      Panel controlPanelButton = new Panel();
      controlPanelButton.setLayout(new FlowLayout());
      controlPanelButton.add(einfuegenOk);
      controlPanelButton.add(loeschenOk);
      controlPanelButton.add(clearTextfield);

      Panel controlPanelFigures = new Panel();
      controlPanelFigures.setLayout(new FlowLayout());
      controlPanelFigures.add(figure[0]);controlPanelFigures.add(figure[1]);
      controlPanelFigures.add(figure[2]);controlPanelFigures.add(figure[3]);
      controlPanelFigures.add(figure[4]);controlPanelFigures.add(figure[5]);
      controlPanelFigures.add(figure[6]);controlPanelFigures.add(figure[7]);
      controlPanelFigures.add(figure[8]);controlPanelFigures.add(figure[9]);

      Panel controlPanelInput = new Panel();
      controlPanelInput.setLayout(new GridLayout(2,1));
      controlPanelInput.add(controlPanelFigures);
      controlPanelInput.add(textfield);

      //Panel um Einfuegen und Loeschen zu steuern
      controlPanel = new Panel();
      controlPanel.setLayout(new BorderLayout());
      controlPanel.setFont(font);
      controlPanel.add("North",new Label("Knoten im AVL-Baum einfuegen/loeschen :"));
      controlPanel.add("West",new Label("   "));
      controlPanel.add("Center",controlPanelInput);
      controlPanel.add("South",controlPanelButton);
      Panel southPanel = new Panel();
      southPanel.add(stop);
      southPanel.add(run);
      southPanel.add(nextStep);

      //panel with textarea and controlPanel
      northPanel = new Panel();
      northPanel.setLayout(new GridLayout(1,2));
      northPanel.add(textarea);
      northPanel.add(controlPanel);

      //Create Layout for Applet
      this.setLayout(new BorderLayout());
      this.add("North",northPanel);
      this.add("Center",scribble);
      this.add("South",southPanel);
      resize(1000,500);
      pack();
      show();
    }
    public boolean handleEvent(Event event) {
      if (event.target instanceof MenuItem) {
         if (((String)event.arg).equals("Schliessen")) {
            this.hide();
            this.dispose();
         }
         else if (((String)event.arg).equals("Geschwindigkeit...")) {               
            SpeedSelect s = new SpeedSelect(publicClass);
            s.pack();
            s.show();
         }
         else if (((String)event.arg).equals("Ungeschehen")) {
            if (!history.empty()) {
               history.pop(); //oberstes Element wegwerfen
               Stack tmp = new Stack();
               tmp = (Stack)history.clone();
               AVLTree avl = new AVLTree();
               rekonstruiereBaum(tmp,avl);
               Tree = avl;
               VerkListe verkliste = new VerkListe();
               Tree.TreeToStack(Tree.TreeTop,verkliste,Tree.imageMiddle,
                             Tree.topBorder,Tree.xHoch2(Tree.CurrentHeight));
               Show show = new Show(publicClass,verkliste);
               posOld = verkliste.Clone();
               if (history.empty()) ungeschehen.disable();
               if (Tree.TreeTop == null) {
                  baum_neu.disable();
                  history = new Stack();
               }
            }
         }
         else if (((String)event.arg).equals("Neu")) {
            Tree = new AVLTree();
            Show show = new Show(publicClass,null);
            baum_neu.disable();
         }
         else if (((String)event.arg).equals("Demo starten")) {
            Tree = new AVLTree();
            Show show = new Show(publicClass,null);
            Demo demo = new Demo(publicClass,mainApplet);
            demo.start();
         }
         else if (((String)event.arg).equals("Hilfe")) {
            Hilfe hilfe = new Hilfe("GESAMT");
         }
         else if (((String)event.arg).equals("Kurze Einleitung")) {
            Hilfe hilfe = new Hilfe("Kurze Einleitung");
         }
         else if (((String)event.arg).equals("Schluessel einfuegen")) {
            Hilfe hilfe = new Hilfe("Schluessel einfuegen");
         }
         else if (((String)event.arg).equals("Schluessel loeschen")) {
            Hilfe hilfe = new Hilfe("Schluessel loeschen");
         }
         else if (((String)event.arg).equals("Clear")) {
            Hilfe hilfe = new Hilfe("Clear");
         }
         else if (((String)event.arg).equals("Ungeschehen ")) {
            Hilfe hilfe = new Hilfe("Ungeschehen ");
         }

         else if (((String)event.arg).equals("Animationsgeschwindigkeit ")) {
            Hilfe hilfe = new Hilfe("Geschwindigkeit");
         }
         else if (((String)event.arg).equals("AVL-Baum/Neu")) {
            Hilfe hilfe = new Hilfe("AVL-Baum/Neu");
         }
         else if (((String)event.arg).equals("AVL-Baum/Demo starten")) {
            Hilfe hilfe = new Hilfe("AVL-Baum/Demo starten");
         }
         else if (((String)event.arg).equals("Stop")) {
            Hilfe hilfe = new Hilfe("Stop");
         }
         else if (((String)event.arg).equals("Run")) {
            Hilfe hilfe = new Hilfe("Run");
         }
         else if (((String)event.arg).equals("Next Step")) {
            Hilfe hilfe = new Hilfe("Next Step");
         }
      }   
      switch(event.id) {
       case Event.ACTION_EVENT:
         if(event.target == einfuegenOk) {
            try {
               int nummer =Integer.valueOf(textfield.getText()).intValue();
               loeschenOk.disable();
               einfuegenOk.disable();
               addNewElement(nummer,null);
               textfield.setText("");
            }
            catch (NumberFormatException nfe) {
               textarea.appendText("Das ist keine Zahl"+"\n");
            }
         }
         else if(event.target == clearTextfield) {
            textfield.setText("");
         }
         else if(event.target == figure[0]) {
            appendNumber(textfield,0);
         }
         else if(event.target == figure[1]) {
            appendNumber(textfield,1);
         }
         else if(event.target == figure[2]) {
            appendNumber(textfield,2);
         }
         else if(event.target == figure[3]) {
            appendNumber(textfield,3);
         }
         else if(event.target == figure[4]) {
            appendNumber(textfield,4);
         }
         else if(event.target == figure[5]) {
            appendNumber(textfield,5);
         }
         else if(event.target == figure[6]) {
            appendNumber(textfield,6);
         }
         else if(event.target == figure[7]) {
            appendNumber(textfield,7);
         }
         else if(event.target == figure[8]) {
            appendNumber(textfield,8);
         }
         else if(event.target == figure[9]) {
            appendNumber(textfield,9);
         }
         else if(event.target == loeschenOk) {
            try {
               int nummer =Integer.valueOf(textfield.getText()).intValue();
               einfuegenOk.disable();
               loeschenOk.disable();
               deleteElement(nummer,null);
               textfield.setText("");
            }
            catch (NumberFormatException nfe) {
               textarea.appendText("Das ist keine Zahl"+"\n");
            }
         }
         else if(event.target == stop) {
            run.enable();
            nextStep.disable();
            stop.disable();
         }
         else if(event.target == run) {
            stop.enable();
            run.disable();
         }
         else if(event.target == nextStep) {
            run.enable();
            stop.disable();
            nextStep.disable();
         }
         break;
       case Event.WINDOW_DESTROY :
         this.hide();
         this.dispose();
	      break;
       case Event.KEY_PRESS :
         if(event.key>=48 && event.key<=57) {
           int zahl = event.key -48; 
           appendNumber(textfield,zahl);
         }
         else if(event.key == 10) {
            try {
               int nummer =Integer.valueOf(textfield.getText()).intValue();
               loeschenOk.disable();
               einfuegenOk.disable();
               addNewElement(nummer,null);
               textfield.setText("");
            }
            catch (NumberFormatException nfe) {
               textarea.appendText("Das ist keine Zahl"+"\n");
            }

         }
         else if(event.key == 127) {
            try {
               int nummer =Integer.valueOf(textfield.getText()).intValue();
               einfuegenOk.disable();
               loeschenOk.disable();
               deleteElement(nummer,null);
               textfield.setText("");
            }
            catch (NumberFormatException nfe) {
               textarea.appendText("Das ist keine Zahl"+"\n");
            }

         }
         break;
       case Event.MOUSE_DOWN :
         int key = find(scribble.offX,scribble.offY);
         if (key>=0) {
            if(event.clickCount>=2) {
               textfield.setText(""+key);
               textarea.setText("Loesche Schluessel "+key+"\n"+"\n");
               textarea.appendText("Bestaetigung mit <Schluessel loeschen>"+"\n");
            }
            else findInsertKeys(key);
         }
         scribble.offX=-1;
         scribble.offY=-1;
         break;
     case Event.MOUSE_MOVE :
         if(scribble.offX>=0) {
            int schluessel = find(scribble.offX,scribble.offY);
            if(schluessel>=0) { //Mauszeiger befindet sich ueber einem Objekt
               if(schluessel != focusKey){ //war Mauszeiger schon davor Objekt
                  Show s1 = new Show(publicClass,posOld);
                  Show s2 = new Show(publicClass,posOld,schluessel,Color.blue);
                  focusKey = schluessel;
               }
            }
            else { //Mauszeiger befindet sich ueber keinem Knoten-Object
               //war Mauszeiger gerade noch ueber einem Object ?
               if(focusKey>=0) {
                  Show s = new Show(publicClass,posOld,focusKey,Color.black);
                  focusKey = -1;
               }
            }
            scribble.offX=-1;
            scribble.offY=-1;
         }
         break;    
	    default :
	      break;
      }
      return true;
    }
    /**Baut neuen AVLTree auf anhand des Stacks history. Alle Einfuege-
    und Loesch-Aktionen werden archiviert. Wurde eingefuegt wird positive,
    beim Loeschen wird negative Zahl in history angehaengt. So kann
    jederzeit jede getaetigte Aktion ungeschehen gemacht werden.*/
    public void rekonstruiereBaum(Stack stack,AVLTree avltree) {
      int zahl;
      while (!stack.empty()) {
         Integer integer = (Integer)stack.pop();
         zahl = integer.intValue();
         rekonstruiereBaum(stack,avltree);
         if (zahl>=0) {   //>=0 means insert operation
            try {avltree.Insert(zahl);}
            catch (TreeInsertDuplicate e) {textarea.appendText("Element "+e.getMessage()+" bereits im AVL-Baum vorhanden"+"\n");}
         }
         else {      //<0 means delete operation      
            try {avltree.Delete(-zahl);}
            catch (TreeDeleteMissing e) {textarea.appendText("Element "+e.getMessage()+" nicht im AVL-Baum vorhanden"+"\n");}
         }
      }
    }
    /**Haengt im Eingabetextfenster Zahl an.*/
    public void appendNumber(TextField textfield,int append) {
      if (textfield.getText().length() <= 2) {
         textfield.setText(textfield.getText()+append);
      }
      else  {
         textarea.appendText("Maximal 3 Ziffern erlaubt"+"\n");
      }
    }

    /**Sucht angeglicketer Knoten im Animationsfenster im AVL-Baum.
    Anhand von der Position eines Mausklicks kann herausgefunden welcher
    Knoten angeglickt wurde. Moeglicherweise befand sich beim Mausklick
    Zeiger ueber keinem Knoten. In diesem Fall wird -1 zurueckgegeben
    andernfalls der Schluesselwert.*/
    //if not found -1 will be returned
    public int find(int x,int y) {
      if ( posOld!=null ) {
         posOld.GoFirst();
         while (!posOld.OutOfRange()) {
            if (posOld.Getx1() <= x && x <= posOld.Getx1()+30
                && posOld.Gety1() <= y && y <= posOld.Gety1()+30 )
                return posOld.GetData();
            else posOld.GoNext();
         }
         return -1;
      }
      return -1;
    }
    /**Findet heraus welche Schluessel links und rechts an einem Knoten (nicht innerer) angehaengt werden kann.*/
    public void findInsertKeys(int key) {
      TreeNode Node;
      Node = Tree.TreeTop;
      if (Tree.TreeTop != null) {
         int left = 0;
         int right = 999;
         while (key != Node.Data) {
            if (key > Node.Data) {
               left = Node.Data + 1;
               Node = Node.Right;
            }
            else {
               right = Node.Data - 1;
               Node = Node.Left;
            }
         }
         int leftBegin = left;
         int leftEnd = Node.Data - 1;
         int rightBegin = Node.Data + 1;
         int rightEnd = right;
         String rangeLeft ="";
         String rangeRight = "";
         if (leftEnd<=leftBegin) rangeLeft = "keine vorhanden";
            else rangeLeft = leftBegin + " bis " + leftEnd;
         if (rightEnd<=rightBegin) rangeRight = "keine vorhanden";
            else rangeRight = rightBegin + " bis " + rightEnd;
         textarea.setText("");
         textarea.appendText("Anhaengen bei Schluessel  " + key + "\n");
         if (Node.Left == null) {
            textarea.appendText("Potentielle Kanditaten links: ");
            textarea.appendText("   " + rangeLeft + "\n");
         }
         if (Node.Right == null) {
            textarea.appendText("Potentielle Kandiaten rechts: ");
            textarea.appendText("   " + rangeRight + "\n");
         }
         if (Node.Right != null && Node.Left != null) {
            textarea.appendText("Innerer Knoten (hat schon zwei Soehne) => Anhaengen nicht moeglich");
         }
      }
    }
    /**Fuegt im AVL-Baum ein neues Element ein und fuehrt alle Klassenaufrufe aus,die dies optisch veranschaulichen.*/
    public void addNewElement (int zahl,Thread thread) {
      try { Tree.Insert(zahl);
         history.push(new Integer(zahl));
         stop.enable();
         run.disable();
         textarea.setText("");
         textarea.appendText("Vorhandener Knoten "+zahl+" im AVL-Baum einfuegen"+"\n");
         textarea.appendText("Regel: Knoten < aktueller Knoten => linker Sohn"+"\n");
         textarea.appendText("       Knoten > aktueller Knoten => rechter Sohn"+"\n");
         textarea.appendText("       bis gesuchter Knoten gefunden"+"\n");
         textarea.appendText("\n");
         VerkListe en = Tree.suchPfad;
         VerkListe in = Tree.initPos;
         VerkListe ln = Tree.leftRot;
         VerkListe rn = Tree.rightRot;
         VerkListe old = Tree.oldPos;
         scribble.hbar.setValue((imageWidth/2)-(scribble.hbar.getVisible()/2));
         switch (Tree.action) {
            case 0: //none
                 //if(posOld!=null) undo = posOld.Clone();
                 ln = null;rn=null;
                 Move move0 = new Move(publicClass,old,in,ln,rn,en,thread);
                 move0.start();
                 posOld = in.Clone();
                 break;
            case 1: //rechts
                 //if(posOld!=null) undo = posOld.Clone();
                 ln = null;
                 Move move1 = new Move(publicClass,old,in,rn,ln,en,thread);
                 move1.start();
                 posOld = rn.Clone();
                 break;
            case 2: //links
                 //if(posOld!=null) undo = posOld.Clone();
                 rn = null;
                 Move move2 = new Move(publicClass,old,in,ln,rn,en,thread);
                 move2.start();
                 posOld = ln.Clone();
                 break;
            case 3: //Rechts-Linksrotation
                 //if(posOld!=null) undo = posOld.Clone();
                 Move move3 = new Move(publicClass,old,in,rn,ln,en,thread);
                 move3.start();
                 posOld = ln.Clone();
                 break; 
            case 4: //Links-Rechtsrotation
                 //if(posOld!=null) undo = posOld.Clone();
                 Move move4 = new Move(publicClass,old,in,ln,rn,en,thread);
                 move4.start();
                 posOld = rn.Clone();
                 break; 
         }
         posOld = new VerkListe();
         Tree.TreeToStack(Tree.TreeTop,posOld,Tree.imageMiddle,Tree.topBorder,Tree.xHoch2(Tree.CurrentHeight));
         ungeschehen.enable();
         baum_neu.enable();
      }
      catch (TreeInsertDuplicate e) {
         einfuegenOk.enable();
         loeschenOk.enable();
         textarea.appendText("Element "+e.getMessage()+" bereits im AVL-Baum vorhanden"+"\n");
      }

    }
    /**Loescht im AVL-Baum ein Element und fuehrt alle Klassenaufrufe aus,die dies optisch veranschaulichen.*/
    public void deleteElement(int zahl,Thread thread) {
      try {
         Tree.Delete(zahl);
         history.push(new Integer(-zahl));
         stop.enable();
         //stopRunCardLayout.show(stopRunPanel,"stop");
         run.disable();
         einfuegenOk.disable();
         loeschenOk.disable();
         textarea.setText("");
         textarea.appendText("Vorhandener Knoten "+zahl+" im AVL-Baum loeschen"+"\n");
         textarea.appendText("\n");
         VerkListe en = Tree.suchPfad;
         VerkListe in = Tree.initPos;
         VerkListe old = Tree.oldPos;
         scribble.hbar.setValue((imageWidth/2)-(scribble.hbar.getVisible()/2));
         MoveDelete move = new MoveDelete(publicClass,old,in,en,zahl,thread);
         move.start();
         if(Tree.TreeTop==null) baum_neu.disable();
      }
      catch (TreeDeleteMissing e) {
         einfuegenOk.enable();
         loeschenOk.enable();
         textarea.appendText("Element "+e.getMessage()+" im AVL-Baum nicht vorhanden"+"\n");
      }
    }
}
       
/**AVL-Tree, Hoehenbalanzierte Binaerbaum.*/
class AVLTree extends Object {
   /**Wurzel des AVL-Baums*/
   public TreeNode TreeTop;
   /**Aktueller Knoten des AVL-Baums*/
   public TreeNode CurNode;
   /**AVL-Baum-Daten werden hier abgelegt.
   Beim Einfuegevorgang: AVL-Baum-Daten (vor Rotation).
   Beim Loeschevorgang: Diverse verk. Liste getrennt durch Erkennungszeichen.*/
   public VerkListe initPos;
   /**AVL-Baum-Daten nach Linksrotation (nur beim Einfuegen) werden hier abgelegt.*/
   public VerkListe leftRot;
   /**AVL-Baum-Daten nach Rechtsrotation (nur beim Einfuegen) werden hier abgelegt.*/
   public VerkListe rightRot;
   /**Einfuegepfad-Daten werden hier abgelegt. Um spaeter bei der optischen
   Veraunschauung der Einfuegebewegung die Koordinaten zu haben.*/
   public VerkListe suchPfad;
   /**AVL-Baum-Daten werden hier abgelegt.*/
   public VerkListe oldPos;
   /**Zeigt an, welche Rotation durchgefuehrt wurde nach Einfuegen oder Loeschen.
   0=keine/1=L/2=R/3=LR/4=RL - Rotation.*/
   public int action;
   /**Aktuelle Hoehe des Baumes.*/
   public int CurrentHeight;
   /**Radius eines Knotens.*/
   public int radius;
   /**Abstand vom obigen Rand.*/
   public int topBorder;
   /**Abstand zwischen jeder Baumebene.*/
   public int yDelta;
   public int imageMiddle = 1300;
   //*********************************************************************
   /**AVL Baum intialisieren.*/
   //*********************************************************************
   public AVLTree() {
      TreeTop = null;
      CurNode = null;
      action = 0;
      CurrentHeight=0;
      radius = 15;
      topBorder = 25;
      yDelta = 4 * radius;
   };

   //****************************************************************
   /**Links-Rotation durchfuehren.*/
   //******************************************************************
   private void Rotate_Right_Child_Up(TreeNode Node, TreeNode NextNode)
   {
      TreeNode TempLowNode = NextNode.Left;
      TreeNode TempHighNode = Node.Up;
      // Wenn Konten nicht Wurzel war,  dann update Elternteil
      if (TempHighNode != null) {
         if (TempHighNode.Left == Node)
            TempHighNode.Left = NextNode;
         else
            TempHighNode.Right = NextNode;
      }

      NextNode.Up = TempHighNode;
      NextNode.Left = Node;

      Node.Up = NextNode;
      Node.Right = TempLowNode;

      // Wenn linkes Kind <> null , dann update Elternteil-Zeiger
      if (TempLowNode != null)
         TempLowNode.Up = Node;

      // Update TreeTop wenn notwendig
      if (Node == TreeTop)
         TreeTop = NextNode;
      int h = height(TreeTop);
      if (CurrentHeight > h) h = CurrentHeight;
         else CurrentHeight = h;
      rightRot = new VerkListe();
      TreeToStack(TreeTop,rightRot,imageMiddle,topBorder,xHoch2(h));
   }

   //************************************************************************
   /**Rechts-Rotation durchfuehren.*/
   //************************************************************************
   private void Rotate_Left_Child_Up(TreeNode Node, TreeNode NextNode)
   {
      TreeNode TempLowNode = NextNode.Right;
      TreeNode TempHighNode = Node.Up;

      /* Wenn Knoten nicht Wurzel war, dann update Elternteil. */
      if (TempHighNode != null) {

         if (TempHighNode.Left == Node)
            TempHighNode.Left = NextNode;
         else
            TempHighNode.Right = NextNode;
      }

      NextNode.Up = TempHighNode;
      NextNode.Right = Node;

      Node.Up = NextNode;
      Node.Left = TempLowNode;

      // Wenn rechtes Kind <>null, dann update Elternteil-Zeiger
      if (TempLowNode != null)
         TempLowNode.Up = Node;

      // Update TreeTop wenn notwendig
      if (Node == TreeTop)
         TreeTop = NextNode;
      int h = height(TreeTop);
      if (CurrentHeight > h) h = CurrentHeight;
         else CurrentHeight = h;
      leftRot = new VerkListe();
      TreeToStack(TreeTop,leftRot,imageMiddle,topBorder,xHoch2(h));
   }
   //**********************************************************************
   /**Hoehe eines Baums oder Teilbaums ermitteln.*/
   //**********************************************************************
   public int height(TreeNode Node) {
      int lh,rh;

      if (Node.Left == null) lh = -1;
      else lh = height(Node.Left);

      if (Node.Right == null) rh = -1;
      else rh = height(Node.Right);

      if (rh > lh) return rh + 1;
      else return lh + 1;
   }

   //************************************************************************
   /**Schluessel in AVL-Baum einfuegen.*/
   //************************************************************************
   public int Insert(int Data) throws TreeInsertDuplicate {
      TreeNode Node, NextNode, ThirdNode;
      // wenn Baum leer dann erstelle einfachen Knoden als Baum
      action = 0; //Voreinstellung keine Aktion
      if (TreeTop == null) {
         TreeTop = new TreeNode(null, Data);
         int h = height(TreeTop);
         initPos = new VerkListe();
         suchPfad = null;
         TreeToStack(TreeTop,initPos,imageMiddle,topBorder,xHoch2(h));
         return Data;
      }
      else {
         // Finde Einfuegestelle und fuege ein
         oldPos = new VerkListe();
         TreeToStack(TreeTop,oldPos,imageMiddle,topBorder,xHoch2(CurrentHeight));
         suchPfad = new VerkListe();
         Node = TreeTop;
         boolean added = false;
         do {

            if (Data > Node.Data) {
               suchPfad.PushToEnd(Node.Data,0,0,0,-1,-1);
               if (Node.Right != null)
                  Node = Node.Right;
               else {
                  suchPfad.PushToEnd(Data,0,0,0,0,0);
                  Node.Right = new TreeNode(Node, Data);
                  added = true;
                  Node.BalanceFactor++;
               }
            }
            else if (Data < Node.Data) {
               suchPfad.PushToEnd(Node.Data,0,0,0,-1,-1);
               if (Node.Left != null)
                  Node = Node.Left;
               else {
                  suchPfad.PushToEnd(Data,0,0,0,0,0);
                  Node.Left = new TreeNode(Node, Data);
                  added = true;
                  Node.BalanceFactor--;
               }
            }
            else
               throw new TreeInsertDuplicate(""+Data);//Duplicate
         } while (! added);

         int h = height(TreeTop);
         if (CurrentHeight > h) h = CurrentHeight;
            else CurrentHeight = h;
         initPos = new VerkListe();
         TreeToStack(TreeTop,initPos,imageMiddle,topBorder,xHoch2(h));

         suchPfad.GoLast();
         suchPfad.GoPrev();
         if (Node.Left !=null) suchPfad.SetLeft(height(Node.Left)+1);
            else suchPfad.SetLeft(0);
         if (Node.Right !=null) suchPfad.SetRight(height(Node.Right)+1);
            else suchPfad.SetRight(0);

         // Update den  Balance-Factor bis eine Rotation notwendig wird
         // oder Wurzel erreicht ist
         while ((Node.Up != null) &&
                ((Node.BalanceFactor == -1) ||
                 (Node.BalanceFactor == 1))) {
            NextNode = Node.Up;
            suchPfad.GoPrev();
            if (NextNode.Left !=null) suchPfad.SetLeft(height(NextNode.Left)+1);
               else suchPfad.SetLeft(0);
            if (NextNode.Right !=null) suchPfad.SetRight(height(NextNode.Right)+1);
               else suchPfad.SetRight(0);
            if (NextNode.Right == Node)
               NextNode.BalanceFactor++;
            else
               NextNode.BalanceFactor--;
            Node = NextNode;
         }

         writeCoord(initPos,suchPfad);
         correctLastBalance(initPos,suchPfad);

         // Wenn  Nicht( |BalanceFactor| <=1)  dann rotiere
         if ((Node.BalanceFactor == 2) || (Node.BalanceFactor == -2)) {

            if (Node.BalanceFactor == 2) {
               NextNode = Node.Right;
               if (NextNode.BalanceFactor == 1) {
                  // Links-Rotation 

                  Node.BalanceFactor = 0;
                  NextNode.BalanceFactor = 0;
                  action = 1;
                  Rotate_Right_Child_Up(Node, NextNode);
               }
               else {
                  // Kind Balance-Factor muss  -1 sein
                  ThirdNode = NextNode.Left;


                  action = 4;
                  if (ThirdNode.BalanceFactor == 1) {
                     Node.BalanceFactor = -1;
                     NextNode.BalanceFactor = 0;
                  }
                  else if (ThirdNode.BalanceFactor == -1) {
                     Node.BalanceFactor = 0;
                     NextNode.BalanceFactor = 1;
                  }
                  else {
                     Node.BalanceFactor = 0;
                     NextNode.BalanceFactor = 0;
                  }
                  ThirdNode.BalanceFactor = 0;
                  Rotate_Left_Child_Up(NextNode, ThirdNode);
                  Rotate_Right_Child_Up(Node, ThirdNode);
               }
            }
            else {
               // Balance-Factor muss  -2 sein
               NextNode = Node.Left;
               if (NextNode.BalanceFactor == -1) {
                  // Rechts-Rotation
                  //Rotate_Left_Child_Up(Node, NextNode);
                  Node.BalanceFactor = 0;
                  NextNode.BalanceFactor = 0;
                  action = 2;
                  Rotate_Left_Child_Up(Node, NextNode);
               }
               else {
                  // Kind Balance-Factor muss 1 sein
                  ThirdNode = NextNode.Right;


                  action = 3;
                  if (ThirdNode.BalanceFactor == -1) {
                     Node.BalanceFactor = 1;
                     NextNode.BalanceFactor = 0;
                  }
                  else if (ThirdNode.BalanceFactor == 1) {
                     Node.BalanceFactor = 0;
                     NextNode.BalanceFactor = -1;
                  }
                  else {
                     Node.BalanceFactor = 0;
                     NextNode.BalanceFactor = 0;
                  }
                  ThirdNode.BalanceFactor = 0;
                  Rotate_Right_Child_Up(NextNode, ThirdNode);
                  Rotate_Left_Child_Up(Node, ThirdNode);
               }
            }
         }
      }
      return Data;
   }
   //************************************************************************
   /**Schluessel im AVL-Baum loeschen.*/
   //************************************************************************
   public void Delete(int Data) throws TreeDeleteMissing {
      TreeNode Node, NextNode, ThirdNode,TmpNode;
      action = 0; //Voreinstellung keine action
      suchPfad = new VerkListe();
      if (TreeTop == null) {
          suchPfad = null;
          throw new TreeDeleteMissing(""+Data);//kein solcher Schlssel vorhanden
      }
      else {
         Node = TreeTop;
         oldPos = new VerkListe();
         TreeToStack(TreeTop,oldPos,imageMiddle,topBorder,xHoch2(CurrentHeight));
         int innererKnoten=-1; //merke innerer Knoten
         boolean deleted = false;
         do {
            if (Data>Node.Data) {
               if (Node.Right != null) {
                  suchPfad.PushToEnd(Node.Data,0,0,0,0,0);
                  Node = Node.Right;
               }
               else throw new TreeDeleteMissing(""+Data);
            }
            else if (Data < Node.Data) {
               if (Node.Left != null) {
                  suchPfad.PushToEnd(Node.Data,0,0,0,0,0);
                  Node = Node.Left;
               }
               else throw new TreeDeleteMissing(""+Data);
            }
            else { //Schluessel gefunden
               if (Node.Left != null && Node.Right != null) { //innerer Knoten
                  TmpNode = getLargestKeyInLeftTree(Node); 
                  Node.Data = TmpNode.Data; //Schluessel vertauschen
                  innererKnoten = TmpNode.Data; //innerer Knoten merken
                  TmpNode.Data = Data;      //EndKnoten bekommt ges. Schluessel
                  suchPfad.PushToEnd(Node.Data,-1,0,0,1,1);
                  Node = Node.Left;
               }
               else if(Node.Left !=null && Node.Right==null) { //Knoten hat einen li. Sohn
                  TmpNode = Node.Left; //gehe zum Blatt
                  suchPfad.PushToEnd(Node.Data,-2,0,0,TmpNode.Data,0);
                  //suchPfad.PushToEnd(TmpNode.Data,0,0,0,0,0);
                  writeCoord(oldPos,suchPfad);
                  Node.Data = TmpNode.Data; //Node:Werte vom Blatt bernehemen
                  Node.BalanceFactor++;
                  Node.Left = null; //altes Blatt abspalten
                  CurNode = Node;
                  deleted = true;
               }
               else if(Node.Right !=null && Node.Left==null) { //Knoten hat einen re. Sohn
                  TmpNode = Node.Right;
                  suchPfad.PushToEnd(Node.Data,-3,0,0,0,TmpNode.Data);
                  //suchPfad.PushToEnd(TmpNode.Data,0,0,0,0,0);
                  writeCoord(oldPos,suchPfad);
                  Node.Data = TmpNode.Data;
                  Node.BalanceFactor--;
                  Node.Right = null;
                  CurNode = Node;
                  deleted = true;
               }
               else if(Node.Right==null && Node.Left==null) { //gesuchter Schluessel ist Blatt
                  deleted = true;
                  suchPfad.PushToEnd(Data,-4,0,0,0,0);
                  writeCoord(oldPos,suchPfad);
                  if (Node.Up==null) { //ges. S. ist Wurzel
                     TreeTop = null;
                     CurNode = TreeTop;
                  }
                  else {
                     TmpNode = Node.Up; //gehe zum Vater
                     if (TmpNode.Right == Node) { //ist Blatt li.oder recht.Sohn vom Vater
                        TmpNode.Right = null;
                        TmpNode.BalanceFactor--;
                     }
                     else {
                        TmpNode.Left = null;
                        TmpNode.BalanceFactor++;
                     }
                     CurNode = TmpNode;
                  }
               }

            }         
         } while (! deleted);  //do-while


         if (innererKnoten >=0) {
            searchInStack(suchPfad,Data);
            int x1tmp = suchPfad.Getx1();
            int y1tmp = suchPfad.Gety1();
            searchInStack(suchPfad,innererKnoten);
            int x2tmp = suchPfad.Getx1();
            int y2tmp = suchPfad.Gety1();
            int tmp = suchPfad.GetData();
            suchPfad.SetData(-1);
            searchInStack(suchPfad,Data);
            suchPfad.Setx1(x2tmp);
            suchPfad.Sety1(y2tmp);
            suchPfad.SetData(tmp);
            searchInStack(suchPfad,-1);
            suchPfad.Setx1(x1tmp);
            suchPfad.Sety1(y1tmp);
            suchPfad.SetData(Data);
         }

         //Update BalanceFaktoren entlag des Suchpfades bis zur Wurzel
         //Rotiere wenn notwendig
         initPos = new VerkListe();
         Node = CurNode;
         while (Node!=null &&
               !(Node.BalanceFactor==-1 || Node.BalanceFactor==1)) {
            // Wenn  Nicht( |BalanceFactor| <=1)  dann rotiere
            action = 0; //Voreinstellung keine action            
            if ((Node.BalanceFactor == 2) || (Node.BalanceFactor == -2)) {
               if (Node.BalanceFactor == 2) {
                  NextNode = Node.Right;
                  if (NextNode.BalanceFactor==1) {
                     // Links-Rotation
                     action = 1;
                     addTreeToStack();
                     Node.BalanceFactor = 0;
                     NextNode.BalanceFactor = 0;
                     Rotate_Right_Child_Up(Node, NextNode);
                     Node = NextNode;
                     initPos.PushToEnd(-1,action,0,0,0,0);
                     initPos.Add(rightRot);
                     rightRot = null;
                  }
                  else if(NextNode.BalanceFactor==0) {
                     // Links-Rotation
                     action = 1;                     
                     addTreeToStack();
                     Node.BalanceFactor = 1;
                     NextNode.BalanceFactor = -1;
                     Rotate_Right_Child_Up(Node, NextNode);
                     Node = NextNode;
                     initPos.PushToEnd(-1,action,0,0,0,0);
                     initPos.Add(rightRot);
                     rightRot = null;
                  }
                  else {
                     // Kind Balance-Factor muss  -1 sein ***rechts-links
                     action = 4; //Rechts-Links-Rot.
                     addTreeToStack();
                     ThirdNode = NextNode.Left;
                     if (ThirdNode.BalanceFactor == 1) {
                        Node.BalanceFactor = -1;
                        NextNode.BalanceFactor = 0;
                     }
                     else if (ThirdNode.BalanceFactor == -1) {
                        Node.BalanceFactor = 0;
                        NextNode.BalanceFactor = 1;
                     }
                     else {
                        Node.BalanceFactor = 0;
                        NextNode.BalanceFactor = 0;
                     }     
                     ThirdNode.BalanceFactor = 0;
                     Rotate_Left_Child_Up(NextNode, ThirdNode);
                     Rotate_Right_Child_Up(Node, ThirdNode);
                     Node = ThirdNode;
                     initPos.PushToEnd(-1,action,0,0,0,0);
                     initPos.Add(leftRot);
                     leftRot = null;
                     initPos.PushToEnd(-1,action,0,0,0,0);
                     initPos.Add(rightRot);
                     rightRot = null;
                  }
               }
               else {
                  // Balance-Factor muss  -2 sein
                  NextNode = Node.Left;
                  if (NextNode.BalanceFactor==-1) {
                     // Rechts-Rotation
                     action = 2;
                     addTreeToStack();
                     Node.BalanceFactor = 0;
                     NextNode.BalanceFactor = 0;
                     Rotate_Left_Child_Up(Node, NextNode);
                     Node = NextNode;
                     initPos.PushToEnd(-1,action,0,0,0,0);
                     initPos.Add(leftRot);
                     leftRot = null;
                  }
                  else if(NextNode.BalanceFactor==0) {
                     // Rechts-Rotation
                     action = 2;
                     addTreeToStack();
                     Node.BalanceFactor = -1;
                     NextNode.BalanceFactor = 1;
                     Rotate_Left_Child_Up(Node, NextNode);
                     Node = NextNode;
                     initPos.PushToEnd(-1,action,0,0,0,0);
                     initPos.Add(leftRot);
                     leftRot = null;
                  }
                  else {
                     // Kind Balance-Factor muss 1 sein ***links-rechts
                     action = 3;
                     addTreeToStack();
                     ThirdNode = NextNode.Right;
                     if (ThirdNode.BalanceFactor == -1) {
                        Node.BalanceFactor = 1;
                        NextNode.BalanceFactor = 0;
                     }
                     else if (ThirdNode.BalanceFactor == 1) {
                        Node.BalanceFactor = 0;
                        NextNode.BalanceFactor = -1;
                     }
                     else {
                        Node.BalanceFactor = 0;
                        NextNode.BalanceFactor = 0;
                     }
                     ThirdNode.BalanceFactor = 0;
                     Rotate_Right_Child_Up(NextNode, ThirdNode);
                     Rotate_Left_Child_Up(Node, ThirdNode);
                     Node = ThirdNode;
                     initPos.PushToEnd(-1,action,0,0,0,0);
                     initPos.Add(rightRot);
                     rightRot = null;
                     initPos.PushToEnd(-1,action,0,0,0,0);
                     initPos.Add(leftRot);
                     leftRot=null;
                  }
               } //***pal(p)==+/- 2***
            }
            if (Node.Up !=null) {
               TmpNode = Node.Up;
               if (TmpNode.Right == Node) {
                  TmpNode.BalanceFactor--;
               }
               else {
                  TmpNode.BalanceFactor++;
               }
            }
            Node = Node.Up;
         } //***while Node!=null***

         if (TreeTop == null) initPos = null; //leerer Baum
            else addTreeToStack();

      } //***TreeTop!=null***
   } //***method delete***
   /**Haelt aktuelle AVL-Baumdaten fest und und haengt dann diese Daten an die verk. Liste initpos an.
   Trennzeichen ist Element mit data=-1.*/
   private void addTreeToStack() {
      VerkListe Tmp = new VerkListe();
      initPos.PushToEnd(-1,action,0,0,0,0);
      TreeToStack(TreeTop,Tmp,imageMiddle,topBorder,xHoch2(CurrentHeight));
      initPos.Add(Tmp);
   }
   /**Clont einen AVL-Baum*/
   public TreeNode Clone(TreeNode oldNode, TreeNode newNode) {
      TreeNode Top = null;
      if(oldNode != null) {
         TreeNode tmp;
         if (newNode==null) {
            tmp = new TreeNode(null,oldNode.BalanceFactor);
            tmp.Data = oldNode.Data;
            Top = tmp;
         }
         else {
            tmp = new TreeNode(newNode,oldNode.BalanceFactor);
            tmp.Data = oldNode.Data;
            if (tmp.Data < newNode.Data) newNode.Left = tmp;
               else newNode.Right = tmp;
         }
         Clone(oldNode.Left,tmp);
         Clone(oldNode.Right,tmp);
      }
      return Top;
   }
   /**Wandelt VerkListe mit AVL-Baum-Daten in AVLTree um.*/
   public AVLTree StackToTree(VerkListe stack) {
      AVLTree AVLtmp = new AVLTree();
      AVLtmp.CurrentHeight = CurrentHeight;
      int treeHeight = CurrentHeight;
      //boolean found = false;
      //tmp.TreeTop = TreeTop;
      if (foundX(stack,imageMiddle)) {
         int data = stack.GetData();
         int bal = stack.GetBalance();
         AVLtmp.TreeTop = new TreeNode(null,data);
         AVLtmp.TreeTop.BalanceFactor = bal;
      }
      for (int h = 1;h<=treeHeight;h++) {
         //int ytmp = topBorder + yDelta * h;
         int restTreeHeight = treeHeight - h + 1; 
         int distance=xHoch2(restTreeHeight)*10; //Abstand in x zum li. und rechten unteren Knoten
         int linkesterKnoten = 0;
         for(int l=treeHeight-restTreeHeight+1;l<=h;l++)
            {linkesterKnoten=linkesterKnoten+xHoch2(treeHeight-l+1)*10;}
         linkesterKnoten = imageMiddle - linkesterKnoten;
         int untererKnoten = linkesterKnoten;
         for(int paare = 1;paare<=h;paare++) {
            //untererKnoten = linkesterKnoten; //linker Sohn
            if(foundX(stack,untererKnoten)){
               int data = stack.GetData();
               int bal = stack.GetBalance();
               if(foundX(stack,untererKnoten+distance)) {//suche oberer Knoten
                  int dataUp = stack.GetData();
                  TreeNode tmp = AVLtmp.searchInTree(dataUp);
                  TreeNode newNode = new TreeNode(tmp,data);
                  newNode.BalanceFactor = bal;
                  tmp.Left = newNode;
               }
            }
            untererKnoten = untererKnoten + distance + distance; //rechter Sohn
            if(foundX(stack,untererKnoten)){
               int data = stack.GetData();
               int bal = stack.GetBalance();
               if (foundX(stack,untererKnoten-distance)) {//suche obererKnoten
                  int dataUp = stack.GetData();
                  TreeNode tmp = AVLtmp.searchInTree(dataUp);
                  TreeNode newNode = new TreeNode(tmp,data);
                  newNode.BalanceFactor = bal;
                  tmp.Right = newNode;
               }
            }
            untererKnoten = untererKnoten + 40; //zum naechsten Paar
         }
      }
      //AVLtmp.CurrentHeight = AVLtmp.height(AVLtmp.TreeTop);
      return AVLtmp;           
   }
   /**Untersuche, ob bestimmte x-Koordinate in VerkListe vorhanden ist.*/
   public boolean foundX(VerkListe stack, int x) {
      boolean found = false;
      stack.GoFirst();
      while (!stack.OutOfRange() && !found) {
         if(x==stack.Getx1()) found = true;
            else stack.GoNext();
      }
      return found;
   }
   /**Sucht Knoten mit groesstem Schluessel im linken Teilbaum.*/
   public TreeNode getLargestKeyInLeftTree(TreeNode Node) {
      Node = Node.Left;
      while (Node.Right != null) {
         Node = Node.Right;
      }
      return Node;
   }
   /**Sucht Knoten mit kleinstem Schluessel im rechten Teilbaum.*/
   public TreeNode getSmallestKeyInRightTree(TreeNode Node) {
      Node = Node.Left;
      while (Node.Left != null) {
         Node = Node.Left;
      }
      return Node;
   }               
   public int xHoch2(int x ) {
      if ( x==0 ) {
         return 0;
      } else {
         int y=1;
         for ( int i=0;i<x ;i++ ) {
          y=y*2;
         }
         return y;
      }
         
   }
   //*********************************************************************
   /**AVL-Baum-Daten werden in einer verketteten Liste abgelegt.
   Durch Travesierung der Knoten (In-Order-Regel  LWR) wird von jedem Knoten
   folgende Daten in verk.Liste abgelegt: data(=Schlussel),
   balance(=Balancefaktor), x1(=x-Koordinate), y1(=y-Koordinate),
   left(=Hoehe linker Teilbaum) und right(=Hoehe rechter Teilbaum)*/
   //*********************************************************************
   public void TreeToStack (TreeNode Node,VerkListe stack,
                            int x, int y, int hoehe) {

      if (Node != null) {
       TreeToStack(Node.Left,stack,x-(hoehe*10),y+yDelta,hoehe/2);

       stack.PushToEnd(Node.Data,Node.BalanceFactor,x,y,0,0);
       if ( Node.Left != null ) stack.SetLeft(height(Node.Left)+1);
       if ( Node.Right != null) stack.SetRight(height(Node.Right)+1);

       TreeToStack(Node.Right,stack,x+(hoehe*10),y+yDelta,hoehe/2);
      }
   }
   //*******************************************************************
   /**Koordinaten in verk. Liste suchPfad eintragen*/
   //******************************************************************
   public void writeCoord(VerkListe from, VerkListe to) {
      to.GoFirst();
      while( ! to.OutOfRange()) {
         int data = to.GetData();
         searchInStack(from,data);
         to.Setx1(from.Getx1());
         to.Sety1(from.Gety1());
         to.GoNext();
      }
   }
   public void correctLastBalance(VerkListe from,VerkListe to) {      
      //System.out.println("------------------");
      //Balance-Faktor eines Knoten im Stack noch berichtigen
      //Vorgaenger-Knoten des neuen Knoten im neuen Baum
      to.GoLast();
      int dataNew = to.GetData();
      to.GoPrev();
      int dataPrev = to.GetData();
      searchInStack(from,dataPrev);
      int balance = from.GetBalance();
      if (dataPrev > dataNew) balance++;
         else balance--;
      from.SetBalance(balance);
   }
   //*********************************************************************
   /**Schluessel number in in verketteten Liste suchen*/
   //*******************************************************************
   //find place of node number
   public void searchInStack  (VerkListe stack,int number) {
      boolean found = false;
      stack.GoFirst();
      while(! stack.OutOfRange() && ! found) {
         if (number == stack.GetData()) found = true;
            else stack.GoNext();
      }
   }
   /** Knoten im AVL-Baum suchen mit Schluessel x*/
   public TreeNode searchInTree (int x) {
      TreeNode Node = TreeTop;
      if (Node != null) {
         if(x>Node.Data ) {
            if (Node.Right!=null) Node = Node.Right;
               else return null;
         }
         else if(x<Node.Data) {
            if (Node.Left!=null) Node = Node.Left;
               else return null;
         }
         else return Node;
      }
      return null;
   }
   /**Gibt Hoehe des Knoten zurueck anhand y-Koordinate.*/
   public int getHeight(int y) {
      y = y-topBorder;
      y = y / yDelta;
      return y;
   }


}
/**Wird ausgeworfen, wenn zu loeschender Schluessel sich nicht im AVL-Baum befindet.*/
class TreeDeleteMissing extends Exception {
   public TreeDeleteMissing(String str) {
      super(str);
   }
}
/**Wird ausgeworfen, wenn einzufuegender Schluessel sich bereits im AVL-Baum befindet.*/
class TreeInsertDuplicate extends Exception {
   public TreeInsertDuplicate(String str) {
      super(str);
   }
}

/**Elemente des AVL-Baums.*/
class TreeNode extends Object {
   public TreeNode Left, Right, Up;
   public int Data;
   public int BalanceFactor;

   //*********************************************************************
   /**Neuen Knoten erstellen.*/
   //*********************************************************************
   public TreeNode(TreeNode Parent, int Data)
   {
      // Initialisierung des neuen Knoten
      Left = null;
      Right = null;
      Up = Parent;
      this.Data = Data;
      BalanceFactor = 0;
   }
}
/**Beidseitige verkettete Liste mit Klasse StackNode als Elemente*/
class VerkListe extends Object {
    public StackNode StackFirst,StackLast,StackTmp,StackCurrent;
    public VerkListe () {
       /**Zeiger auf erstes Element der verk. Liste*/
       StackFirst = null;
       /**Zeiger auf letztes Element der verk. Liste*/
       StackLast = null;
       StackTmp = null;
       /**Zeiger auf aktuelles Element der verk. Liste*/
       StackCurrent = null;
    }
    /**Fuegt neues Element der verk. Liste zu. Wird benutzt um
    AVL-Baum-Daten aufzubewahren. Data=Schluessel, Balance=Balancefaktor, 
    x1=x-Koordinate, y1=y-Koordinate, left=Hoehe linker Teilbaum,
    right=Hoehe rechter Teilbaum. */
    public void PushToEnd(int Data,int Balance,int x1,int y1,int left,
                          int right) {
        if(StackFirst==null) {
            StackFirst = new StackNode(Data,Balance,x1,y1,left,right);
            StackLast = StackFirst;
            StackCurrent = StackFirst;
        }
        else {
            StackTmp = new StackNode(Data,Balance,x1,y1,left,right);
            StackTmp.Prev = StackLast;
            StackLast.Next = StackTmp;
            StackLast = StackTmp;
            StackCurrent = StackTmp;
        }
    }
    /**Fuegt neues Element der verk. Liste zu. Wird benutzt um
    Bewegungs-Daten aufzubewahren.*/
    public void PushToEnd(int Data,int Balance,int x1,float y2,float m,
                          int right) {
        if(StackFirst==null) {
            StackFirst = new StackNode(Data,Balance,x1,y2,m,right);
            StackLast = StackFirst;
            StackCurrent = StackFirst;
        }
        else {
            StackTmp = new StackNode(Data,Balance,x1,y2,m,right);
            StackTmp.Prev = StackLast;
            StackLast.Next = StackTmp;
            StackLast = StackTmp;
            StackCurrent = StackTmp;
        }
    }
    /**Gibt Laenge der verk. Liste zurueck.*/
    public int Length() {
      int z = 0;
      if (StackFirst==null || StackLast==null) return z;
      GoFirst();
      while (! OutOfRange()) {
         z=z+1;
         GoNext();
      }
      return z;
    }
    /**Geht zum x-ten Element der verk. Liste (vom Anfang).*/
    public void Go(int x) {
      if (x > Length()) x = Length();
      if (x < 1) x = 1;
      GoFirst();
      for(int i = 1; i<x;i++) {
         GoNext();
      }
    }
    /**Setzt StackCurrent ein Element weiter in der verk. Liste.*/
    public void GoNext() {StackCurrent = StackCurrent.Next;}
    /**Setzt StackCurrent. ein Element zurueck in der verk. Liste.*/
    public void GoPrev() {StackCurrent = StackCurrent.Prev;}
    public int Getx1() {return StackCurrent.x1;}
    public int Gety1() {return StackCurrent.y1;}
    public int GetLeft() {return StackCurrent.left;}
    public int GetRight() {return StackCurrent.right;}
    public int GetData() {return StackCurrent.Data;}
    public int GetBalance() {return StackCurrent.Balance;}
    public float GetM() {return StackCurrent.m;}
    public float Gety2() {return StackCurrent.y2;}
    public void SetData(int x) {StackCurrent.Data =x;}
    public void SetBalance(int x) {StackCurrent.Balance =x;}
    public void Setx1(int x) {StackCurrent.x1 = x;}
    public void Sety1(int y) {StackCurrent.y1 = y;}
    public void Sety2(float y) {StackCurrent.y2 = y;}
    public void SetLeft(int l) {StackCurrent.left = l;}
    public void SetRight(int r) {StackCurrent.right = r;}
    /**Gibt an, ob aktueller Zeiger sich in oder ausserhalb der Liste sich befindet.*/
    public boolean OutOfRange() {
      if (StackCurrent == null) {
	    return true;
      } else {
         return false;
      }
    }
    /**Geht zum ersten Element der verk. Liste.*/
    public void GoFirst(){
        StackCurrent = StackFirst;
    }
    /**Geht zum letzten Element der verk. Liste.*/
    public void GoLast() {
        StackCurrent = StackLast;
    }
    /**Loescht Element mit data=del aus der verk. Liste.*/
    public void delete(int del) {
      StackNode predecessor,successor;
      search(del);
      predecessor = StackCurrent.Prev;
      successor = StackCurrent.Next;
      if (StackCurrent == StackFirst && StackCurrent == StackLast) {
         StackFirst=null;StackLast=null;StackCurrent=null;
      }
      else if (StackCurrent == StackFirst) {
         GoNext();
         if(StackCurrent != null) {
            StackCurrent.Prev = null;
            StackFirst = StackCurrent;
         }
         else {
            StackFirst=null;StackLast=null;StackCurrent=null;
         }
      }
      else if(StackCurrent == StackLast) {
         GoPrev();
         if(StackCurrent != null) {
            StackCurrent.Next = null;
            StackLast = StackCurrent;
         }
         else {
            StackFirst=null;StackLast=null;StackCurrent=null;
         }
      }
      else {
         predecessor.Next = successor;
         successor.Prev = predecessor;
      }
    }
    public void swapData(int data1,int data2) {
      search(data1);
      StackTmp = StackCurrent;
      search(data2);
      StackCurrent.Data = data1;
      StackTmp.Data = data2;
    }
    /**Setzt StackCurrent auf gesuchtes Element data.*/
    public void search(int data) {
      GoFirst();
      boolean found = false;
      while (!OutOfRange() && !found) {
         if(StackCurrent.Data == data) found = true;
            else GoNext();
      }
    }
    /**Kopiert die verk. Liste.*/  
    public VerkListe Clone() {
      VerkListe tmp = new VerkListe();
      GoFirst();
      while( ! OutOfRange()) {
        tmp.PushToEnd(StackCurrent.Data,StackCurrent.Balance,
        StackCurrent.x1,StackCurrent.y1,
        StackCurrent.left,StackCurrent.right);
        GoNext();
      }
      return tmp;
    }
    /**Haengt eine verk. Liste an.*/
    public void Add(VerkListe stack) {
      //stack.GoFirst();
      StackLast.Next = stack.StackFirst;
      stack.StackFirst.Prev = StackLast;
      StackLast = stack.StackLast;
    }
    /**Trennt zwei verk. Listen.*/
    public void Split(VerkListe stack)  {
     if(stack!=null) {
      if (stack.GetData()<0) stack.GoNext();
      StackNode ende=null;
      StackFirst = stack.StackCurrent;
      StackFirst.Prev = null;
      while ( stack.StackCurrent.Next!=null  &&  stack.GetData()>=0) {
         stack.GoNext();
      }
      if (stack.GetData()<0) {
         ende = stack.StackCurrent.Prev;
         stack.StackCurrent.Prev = null;
         stack.StackFirst = stack.StackCurrent;
      }
      else {
         ende = stack.StackCurrent;
         stack.StackFirst = null;
         stack.StackLast = null;
         stack.StackCurrent = null;
      }
      StackLast = ende;
      StackLast.Next = null;
     }// stack!=null
    }
    /**Finde zu vertauschende Schluessel und erzeuge stack und gebe diesen zurueck.*/
    public VerkListe prepareSwap(int data1,int data2) {
       VerkListe tmp = new VerkListe();
       search(data1);
       tmp.PushToEnd(GetData(),0,Getx1(),Gety1(),0,0);
       search(data2);
       tmp.PushToEnd(GetData(),0,Getx1(),Gety1(),0,0);
       return tmp;
    }

}

/**Elemente der verketteten Liste.*/
class StackNode extends Object {
    StackNode Prev,Next;
    int Data,Balance;
    int x1,y1;
    int left,right;
    float m,y2;
    public StackNode (int Data,int Balance,int x1,int y1,int left,int right) {
        Next = null;
        Prev = null;
        this.Data = Data;
        this.Balance = Balance;
        this.x1 = x1;
        this.y1 = y1;
        this.left = left;
        this.right = right;
        this.m=0;
        this.y2=0;
    }
    public StackNode (int Data,int Balance,int x1,float y2,float m,int right) {
        Next = null;
        Prev = null;
        this.Data = Data;
        this.Balance = Balance;
        this.x1 = x1;
        this.y1 = 0;
        this.left = 0;
        this.right = right;
        this.m = m;
        this.y2 = y2;
    }
}
